"""
Copyright (c) 2021 Synopsys, Inc.
Use subject to the terms and conditions of the Synopsys End User Software License and Maintenance Agreement.
All rights reserved worldwide.
"""

import os
import copy
import unicodedata
from typing import Tuple, List

from urllib.parse import quote, unquote

from ..custom_types import ExtendedFullPath

# from https://github.com/google/guava/blob/master/guava/src/com/google/common/net/UrlEscapers.java
PATH_SEGMENT_ENCODING_SAFE_CHARS = "a-zA-Z0-9-._~!$'()*,;&=@:+"
PATH_FRAGMENT_ENCODING_SAFE_CHARS = PATH_SEGMENT_ENCODING_SAFE_CHARS + "/?"

# translates apples to oranges because fun fun and we need some workarounds for BDIO parser bugs.
file_scheme_translation_table = {
    "file": "unknown",  # avoid file:file... as that makes the parser choke
    "7z": "sevenz",  # schemes can't start with a number - doh
    "": "unknown",  # due to bugs we might get an empty scheme
}


def simple_fullpath_to_extended_fullpath(fullpath: List[str]) -> ExtendedFullPath:
    extended_fullpath = []
    for pathpart in fullpath:
        extended_fullpath.append({"type": "file", "path": pathpart})
    return extended_fullpath


def extended_fullpath_to_simple_fullpath(extended_fullpath: ExtendedFullPath) -> List[str]:
    fullpath = list(map(lambda p: p["path"], extended_fullpath))
    return fullpath


def deduplicate_extended_fullpaths(
    extended_fullpaths: List[ExtendedFullPath],
) -> List[ExtendedFullPath]:
    result: List[ExtendedFullPath] = []
    unique_paths = set()
    for extended_fullpath in extended_fullpaths:
        fullpath = tuple(extended_fullpath_to_simple_fullpath(extended_fullpath))
        if fullpath not in unique_paths:
            unique_paths.add(fullpath)
            result.append(extended_fullpath)

    return result


def fullpath_to_bdio_path_uri(extended_fullpath: ExtendedFullPath) -> str:
    """
    Repeatedly URI encode the path parts and prepend with the scheme/type
    see https://blackducksoftware.github.io/bdio/specification#file-paths
    The implementation of that algo is to be found when following:
    https://github.com/blackducksoftware/bdio/blob/master/bdio2/src/main/java/com/blackducksoftware/bdio2/LegacyScanContainerEmitter.java#L172
    https://github.com/blackducksoftware/magpie-libraries/blob/master/magpie/src/main/java/com/blackducksoftware/common/value/HID.java#L499

    :param extended_fullpath: a findlib results extended fullpath
    :return: a BDIO path URI
    """

    def get_path_parts(fullpath_entry, translate_scheme=True) -> Tuple[str, str]:
        path = fullpath_entry["path"]
        if path.endswith(os.sep):  # no trailing / allowed by the spec
            path = path.rstrip(os.sep)
        # TODO: add/remove prefix '/' according to rules in
        # https://blackducksoftware.github.io/bdio/specification#common-file-path-archive-schemes
        # NOTE that those schemes are incomplete! Would need to clarify further rules.
        path = os.path.join(os.sep, path)
        # the Hub chokes on unicode strings which are not normalized so we need to obey
        path = unicodedata.normalize("NFC", path)

        scheme = fullpath_entry["type"]
        if translate_scheme:
            scheme = file_scheme_translation_table.get(scheme, scheme)
        if scheme[0].isdigit():
            raise ValueError("Path type cannot start with a digit - forbidden by BDIO spec.")

        return scheme, path

    extended_fullpath = copy.deepcopy(extended_fullpath)  # content needs some modifications applied

    # BDIO path encoding wants the scheme to be applied to the children of an archive
    # not to itself, as it tells what extractor _was_ used to get to the child file
    # shifting here all the schemes down to the next path-element
    scheme = "file://"  # first entry must always be of scheme "file"
    for path_extended in extended_fullpath:
        current_scheme = path_extended["type"]
        path_extended["type"] = scheme
        scheme = current_scheme

    # first entry has different building rules
    scheme, path = get_path_parts(extended_fullpath[0], translate_scheme=False)
    # first path must not have `/` escaped
    result = scheme + quote(path, safe=PATH_SEGMENT_ENCODING_SAFE_CHARS + "/")

    for path_extended in extended_fullpath[1:]:
        scheme, path = get_path_parts(path_extended)
        result = (
            scheme
            + ":"
            + quote(result, safe=PATH_SEGMENT_ENCODING_SAFE_CHARS)
            + "#"
            + quote(path, safe=PATH_FRAGMENT_ENCODING_SAFE_CHARS)
        )

    return result


def bdio_path_uri_to_fullpath(path_uri: str) -> ExtendedFullPath:
    """
    Decode the BDIO URI into a findlib extended fullpath
    .
    :param path_uri: the BDIO path URI
    :return: a findlib results style extended fullpath
    """
    path_uri_part = path_uri
    extended_fullpath = []
    max_nested_paths = 1000  # to avoid potential infinite loops on malformed paths
    while max_nested_paths > 0:
        if path_uri_part.startswith("file://"):
            scheme = "file"
            path = path_uri_part.replace("file://", "", 1)
            if path.startswith(os.sep) and len(path) > 1:
                path = path.lstrip(os.sep)
            extended_path_part = {"type": scheme, "path": path}
            extended_fullpath.append(extended_path_part)
            break

        try:
            scheme, rest = path_uri_part.split(":", 1)
            path_uri_part, path = rest.split("#")
        except ValueError:
            # not enough values to unpack, the URI is malformed
            break
        path_uri_part = unquote(path_uri_part)
        path = unquote(path)
        if path.startswith(os.sep):
            path = path.lstrip(os.sep)
        extended_path_part = {"type": scheme, "path": path}
        extended_fullpath.append(extended_path_part)  # this is reverse but will be changed later
        max_nested_paths = max_nested_paths - 1

    # shifting the path schemes back up as the encoder shifts them down
    scheme = "file"  # first entry is of scheme "file" because we dropped it during encoding
    for path_extended in extended_fullpath:
        current_scheme = path_extended["type"]
        path_extended["type"] = scheme
        scheme = current_scheme

    # finally reverse the order as the paths were encoded inside out
    return list(reversed(extended_fullpath))



"""
Copyright (c) 2021 Synopsys, Inc.
Use subject to the terms and conditions of the Synopsys End User Software License and Maintenance Agreement.
All rights reserved worldwide.
"""

import os
from zipfile import ZipFile, ZipInfo


class MyZipFile(ZipFile):

    def extract(self, member, path=None, pwd=None):
        if not isinstance(member, ZipInfo):
            member = self.getinfo(member)

        if path is None:
            path = os.getcwd()

        ret_val = self._extract_member(member, path, pwd)
        attr = member.external_attr >> 16
        if attr != 0:
            os.chmod(ret_val, attr)
        return ret_val

    def extractall(self, path=None, members=None, pwd=None):
        if members is None:
            members = self.namelist()

        if path is None:
            path = os.getcwd()
        else:
            path = os.fspath(path)

        for zipinfo in members:
            self.extract(zipinfo, path, pwd)



"""
Copyright (c) 2021 Synopsys, Inc.
Use subject to the terms and conditions of the Synopsys End User Software License and Maintenance Agreement.
All rights reserved worldwide.
"""

import os
import tqdm
import tarfile
import logging
import pandas as pd
import re
import shutil
import csv
import json
import ast
from blackduck_c_cpp.util.build_log_parser import LogParser
from blackduck_c_cpp.pkg_manager.ld_debug_parse import LdDebugParser
from collections import defaultdict
import hashlib
import sys
import itertools
import subprocess
from blackduck_c_cpp.util import global_settings
from blackduck_c_cpp.util import util
from datetime import datetime

PKG_MGR_MODE = 'pkg_mgr'
BDBA_MODE = 'bdba'
ALL_MODE = 'all'


class PkgManagerBom:
    """
    Runs package manager on all linker,header,executable and transitive dependency files
    and creates a tar file to send to bdba
    """

    def __init__(self, pkg_manager, build_log, cov_home, blackduck_output_dir, os_dist, bld_dir, unresolved_files_list,
                 resolved_files_list, binary_files_list, json_grep_files_list, skip_build, skip_transitives,
                 skip_dynamic, run_modes,
                 cov_header_files, debug, header_files_set, linker_files_set, executable_files_set,
                 resolved_cov_files_list, unresolved_cov_files_list,
                 hub_api, offline, use_offline_files):
        self.pkg_manager = pkg_manager
        self.build_log = build_log
        self.skip_build = skip_build
        self.skip_transitives = skip_transitives
        self.skip_dynamic = skip_dynamic
        self.output_dir = blackduck_output_dir
        self.cov_home = cov_home
        self.header_list = cov_header_files
        self.bld_dir = bld_dir
        self.os_dist = os_dist
        self.resolved_files_path = resolved_files_list
        self.unresolved_files_path = unresolved_files_list
        self.run_modes = run_modes

        self.resolved_cov_files_path = resolved_cov_files_list
        self.unresolved_cov_files_path = unresolved_cov_files_list

        self.binary_files_path = binary_files_list
        self.json_grep_files_list = json_grep_files_list
        self.debug = debug
        self.cov_header_files_set = header_files_set
        self.cov_linker_files_set = linker_files_set
        self.cov_executable_files_set = executable_files_set
        self.hub_api = hub_api
        self.offline = offline
        self.use_offline_files = use_offline_files

        self.csv_file_path = os.path.join(self.output_dir, "raw_bdio.csv")
        self.json_file_path = os.path.join(self.output_dir, "raw_bdio.json")
        self.executable_list = None
        self.linker_list = None
        self.lib_path_set = None
        self.build_dir_dict = None
        self.components_data = None
        self.final_components_dict = None
        self.file_dict = dict()
        self.resolved_file_dict = dict()
        self.unresolved_file_dict = dict()
        self.cov_file_dict = dict()
        self.resolved_cov_file_dict = dict()
        self.unresolved_cov_file_dict = dict()
        self.pkg_mgr_data = dict()
        self.bdba_data = dict()
        self.ldd_linker = dict()
        self.lddebug_executable = dict()
        self.unrecognized_pkg_manager = True
        self.direct_match_evidence = 'Binary'
        self.transitive_match_evidence = 'Binary'
        self.unresol_set = set()

    def run_log_parser(self):
        """
        This function calls LogParser.parse_build_log method to get all paths from linker invocations,
        all output executables and -L paths from build_log.txt for path resolution

        """
        start_parser = datetime.now()
        log_parser = LogParser()
        self.linker_list = log_parser.parse_build_log(self.build_log, self.os_dist)
        self.lib_path_set = log_parser.lib_path
        self.executable_list = log_parser.get_exectuable(self.build_log)
        end_parser = datetime.now()
        logging.info("time taken to parse for files from build-log.txt : {}".format(end_parser - start_parser))

    def get_files(self, list_paths):
        """
        Function to create a dictionary with input list of paths in sorted order with length
        {a.so: [g/b/c/a.so,c/a.so],...}
        param: list of paths
        return: dictionary containing key as basefile and value as list of all paths with basefile
        """
        basefile_list = list(map(lambda path: os.path.basename(path).strip(), list_paths))
        basefiles_dict = dict.fromkeys(basefile_list, None)
        for basefile in basefiles_dict.keys():
            basefiles_dict[basefile] = sorted(
                set(map(lambda path_1: path_1.strip(),
                        filter(lambda path: os.path.basename(path).endswith(basefile), list_paths))),
                key=len,
                reverse=True)
        return basefiles_dict

    def generate_build_dir_files(self):
        """ This function gets all files in build dir
        return: dictionary containing key as basefile and value as list of all paths with basefile in build directory
        """
        start_gen = datetime.now()
        basefiles_dict = dict()
        dirs_set = set()
        count = 0
        for (root, dirs, files) in os.walk(self.bld_dir, followlinks=True):
            logging.debug("root is {}".format(root))
            st = os.stat(root)
            walk_dirs = []
            for dirname in dirs:
                st = os.stat(os.path.join(root, dirname))
                dir_key = st.st_dev, st.st_ino
                if dir_key not in dirs_set:
                    dirs_set.add(dir_key)
                    walk_dirs.append(dirname)
            for file in files:
                if file in self.unresol_set:
                    if file in basefiles_dict:
                        basefiles_dict[file] = basefiles_dict[file] + [os.path.join(root, file)]
                    else:
                        basefiles_dict[file] = [os.path.join(root, file)]
                    count += 1
                    if (count % 1000 == 0):
                        logging.debug("files resolved so far {}".format(len(basefiles_dict.keys())))
            dirs[:] = walk_dirs
        end_gen = datetime.now()
        logging.debug("total list of files found are : {}".format(len(basefiles_dict.keys())))
        logging.debug("time taken to get all files in build dir : {}".format(end_gen - start_gen))
        return basefiles_dict

    def test_for_dash_l_file_completion(self, path_list):
        """ This function will attempt to resolve paths to files based on paths found
        as -L flags in the build-log
        param: list of paths
        return: set of resolved paths with length used for sort in descending order
        """
        resolved_paths = set()
        for path in path_list:
            if os.path.exists(path):
                resolved_paths.add(path)
            else:
                for lib_path in self.lib_path_set:
                    test_path = os.path.abspath(util.resolve_path(os.path.join(lib_path, path)))
                    if os.path.exists(test_path):
                        resolved_paths.add(test_path)
                        break
        return sorted(resolved_paths, key=len, reverse=True)

    def resolve_file_paths(self, basefiles_dict):
        """ This function separates all files to resolved and unresolved files
        param: dictionary containing key as basefile and value as list of all paths with basefile
        return: resolved and unresolved dictionaries with key as basefile and value as list of all paths with basefile
        """
        resolved_files_dict = dict()
        unresolved_files_dict = dict()
        for basefile, paths_list in basefiles_dict.items():
            resolved_paths = self.test_for_dash_l_file_completion(paths_list)
            if len(resolved_paths) > 0:
                resolved_files_dict[basefile] = set(resolved_paths)
            else:
                try:
                    """looks in build directory files"""
                    if basefile in self.build_dir_dict:
                        resolved_files_dict[basefile] = set(self.build_dir_dict[basefile])
                    else:
                        """ add to unresolved dictionary only if not even a single path in paths_list is resolved"""
                        unresolved_files_dict[basefile] = set(paths_list)
                except IndexError:
                    continue
        return resolved_files_dict, unresolved_files_dict

    def write_to_txt(self, input_dict, path):
        """This function is used to write resolved files, unresolved files and all files to text files
        param: dictionary to write to a file
        """
        try:
            with open(path, "w") as file_wr:
                file_wr.write("{\n")
                for type in input_dict.keys():
                    file_wr.write("'For {}:'\n".format(type))
                    for basefile in input_dict[type].keys():
                        file_wr.write("'{}':'{}'\n".format(basefile, input_dict[type][basefile]))
                        file_wr.write("\n")
                file_wr.write("}")
        except OSError:
            logging.error("Couldn't write file: {}".format(path))

    def join_components(self, sha1, confidence, pkg_result, matchType):
        """ This function joins sha1,confidence and package manager result with matchType to return a single string
        param:
            sha1: string
            confidence: string
            pkg_result: string
            matchType: string
        return: string
        """
        matchType = "matchType:{}".format(matchType)
        pkg_result = " ".join([pkg_result, sha1])
        pkg_result = " ".join([pkg_result, confidence])
        pkg_result = " ".join([pkg_result, matchType])
        return pkg_result

    def check_pkg_mgr(self, basefiles_dict):
        """ This function sends files to package manager and if no result adds those files to bdba
        param: dictionary containing key as basefile and value as list of all paths with basefile
        return: package manager dicitonary with key as filepath and value as result of package manager
                bdba dicitonary with key as basefile and value as list of all paths with basefile
        """
        if self.pkg_manager == 'rpm' or self.pkg_manager == 'dpkg':
            self.unrecognized_pkg_manager = False
        pkg_dict = {}
        bdba_dict = {}
        MatchType = self.direct_match_evidence
        pattern_file = re.compile('(?<=file:)([^ ])*')
        for basefile, paths_list in basefiles_dict.items():
            sha1, confidence, pkg_result, result = self.run_pkg_query(paths_list)
            if result:
                pkg_result = self.join_components(sha1, confidence, pkg_result, MatchType)
                file_path = re.search(pattern_file, pkg_result).group()
                pkg_dict[file_path] = pkg_result
                if self.debug or self.unrecognized_pkg_manager:
                    bdba_dict[basefile] = paths_list
            else:
                bdba_dict[basefile] = paths_list
        return pkg_dict, bdba_dict

    def run_pkg_query(self, paths_list):
        """ This function calls respective package manager function
        param: list of paths
        return: returns package manager result if exists otherwise returns "error"
            sha1: string
            confidence: string
            pkg_result: string
            result: Boolean
        """
        result = False
        sha1, confidence, pkg_result = "error", "error", "error"
        for path in paths_list:
            symlink_command = "readlink -f " + "{}".format(path)
            out = subprocess.getstatusoutput(symlink_command)
            status, symlink_path = util.value_getstatusoutput(out)
            if status and ("no path found matching pattern" not in symlink_path):
                sha1, confidence, pkg_result = getattr(self, self.call_func("command"))([path, symlink_path])
            else:
                sha1, confidence, pkg_result = getattr(self, self.call_func("command"))([path])
            if ("\n" not in pkg_result) and ("error" not in pkg_result):
                result = True
                break;
        return sha1, confidence, pkg_result, result

    def run_dpkg_command(self, paths_list):
        """ This function runs dpkg package manager command
        param: list of paths
        return: sha1,confidence and package result - all params in string format
        """
        sha1, confidence, dpkg_result = "error", "error", "error"
        for path in paths_list:
            dpkg_command = "dpkg-query --showformat='file:" + "{}".format(
                path) + " src:${Source}  pkg:${Package} version:${Version} package-architecture:${Architecture} timestamp:${db-fsys:Last-Modified} size:${Installed-Size}\n' --show `dpkg -S " + "{}".format(
                path) + "| awk -F: '{print $1}'`"
            out = subprocess.getstatusoutput(dpkg_command)
            status, dpkg_result_1 = util.value_getstatusoutput(out)
            if status and ("dpkg: error:" not in dpkg_result_1) and (
                    "no path found matching pattern" not in dpkg_result_1):
                confidence = "confidence:100"
                sha1 = self.find_sha1sum(path)
                dpkg_result = dpkg_result_1
                break;
        return sha1, confidence, dpkg_result

    def run_rpm_command(self, paths_list):
        """ This function runs rpm package manager command
        param: list of paths
        return: sha1,confidence and package result - all params in string format
        """
        sha1, confidence, rpm_result = "error", "error", "error"
        for path in paths_list:
            rpm_command = "rpm -q --queryformat " + "'file:{} ".format(path) + \
                          "src:%{SOURCERPM} pkg:%{Name} version:%{Version} package-architecture:/%{ARCH} timestamp:%{FILEMTIMES} size:%{FILESIZES} epoch:%{EPOCH}: release:-%{RELEASE}\n' --whatprovides " \
                          + "{}".format(path)
            out = subprocess.getstatusoutput(rpm_command)
            status, rpm_result_1 = util.value_getstatusoutput(out)
            if status and ("no path found matching pattern" not in rpm_result_1):
                confidence = "confidence:100"
                sha1 = self.find_sha1sum(path)
                rpm_result = rpm_result_1
                break;
        return sha1, confidence, rpm_result

    def run_brew_command(self, paths_list):
        """ This function runs brew package manager command
        param: list of paths
        return: sha1,confidence and package result - all params in string format
        """
        sha1, confidence, brew_result = "error", "error", "error"
        match_type = self.direct_match_evidence
        match_type = "matchType:{}".format(match_type)
        for path in paths_list:
            path_split = path.split('/')
            idx = 0
            for element in path_split:
                idx += 1
                if element == 'Cellar':
                    try:
                        brew_result = "file:{} pkg:{} version:{}".format(path, path_split[idx].split("@")[0],
                                                                         path_split[idx + 1])
                        sha1 = self.find_sha1sum(path)
                        confidence = "confidence:100"
                        size = "size:0"
                        timestamp = "timestamp:111"
                        package_architecture = "package-architecture:none"
                        brew_result = " ".join([brew_result, timestamp])
                        brew_result = " ".join([brew_result, size])
                        brew_result = " ".join([brew_result, package_architecture])
                        brew_result = " ".join([brew_result, match_type])
                        break;
                    except IndexError:
                        break;
        return sha1, confidence, brew_result

    def run_windows_command(self, paths_list):
        """ This function runs for windows or when package manager and os_dist is not found
        """
        return "error", "error", "error"

    def run_not_found_command(self, paths_list):
        """when package manager or os_dist is not found
        """
        return "error", "error", "error"

    def check_ldd(self, basefiles_dict):
        """ This function calls ldd function of all .so files
        param: dictionary containing key as basefile and value as list of all paths with basefile
        return: dictionary with key as filepath and
        value as dictionary with key transitive dependency element and value package manager results for it
        """
        ldd_linker = dict()
        bdba_dict = dict()
        for filepath, pkg_result in basefiles_dict.items():
            if re.match(r".*\.so$|.*\.so\..+$", filepath):
                list_ldd = self.get_ldd_for_sharedlib(filepath)
                if len(list_ldd) > 0:
                    ldd_linker[filepath], bdba_dict = self.check_pkg_for_ldds(list_ldd, bdba_dict)
        return ldd_linker, bdba_dict

    def check_pkg_for_ldds(self, ldd_list, bdba_dict):
        """ This function calls package manager for ldds
        param: list of transitive dependencies for each .so file
        return: dictionary with key transitive dependency element and value as package manager result for it
        """
        match_type = self.transitive_match_evidence
        ldd_dict = dict()
        for ldd_ele in ldd_list:
            ldd_ele = ldd_ele.strip()
            sha1, confidence, pkg_result, result = self.run_pkg_query(
                [ldd_ele, os.path.basename(ldd_ele)])
            if result:
                pkg_result = self.join_components(sha1, confidence, pkg_result, match_type)
                ldd_dict[ldd_ele] = pkg_result
                if self.debug:
                    bdba_dict[os.path.basename(ldd_ele)] = set([ldd_ele])
            else:
                bdba_dict[os.path.basename(ldd_ele)] = set([ldd_ele])
        return ldd_dict, bdba_dict

    def get_ldd_for_sharedlib(self, so_filepath):
        """This function calls ldd parse function on .so files
        param: .so filepath - string
        return: list of distinct transitive dependencies
        """
        if not os.path.exists(so_filepath):
            return []
        output_path_set = self.ldd_output_parse(so_filepath)
        return list(output_path_set)

    def ldd_output_parse(self, full_path):
        """This function performs parsing on output of ldd command
        param: .so filepath - string
        return: list of distinct transitive dependencies
        """
        output_path_set = set()
        ldd_return = subprocess.getoutput('ldd {}'.format(full_path))
        parsed_results = re.findall(r'^\s*(\S+)\s+=>(.*) \(.*\)$', ldd_return, re.MULTILINE)
        for fname, path in parsed_results:
            output_path_set.add(path)
        return output_path_set

    def find_sha1sum(self, path):
        """This function finds sha1 of a file
        param: filepath - string
        return: sha1 - string
        """
        sha1 = hashlib.sha1()
        try:
            with open(path, 'rb') as f:
                data = f.read()
                sha1.update(data)
                sha1_result = "sha1sum:{}".format(sha1.hexdigest())
        except Exception as e:
            # logging.debug("exception is {}".format(e))
            sha1_result = "sha1sum:notAvailable"
        return sha1_result

    def call_func(self, string):
        """To call specific package manager function
        param: part of function name - string
        return: full function name - string
        """
        return "run_" + self.pkg_manager + '_' + string

    def can_perform_ldd(self):
        """This function checks if ldd is present in the system"""
        return shutil.which('ldd') is not None

    def check_ldd_deb(self, executables):
        """This function calls ld debug function on executables
        param: dictionary of executables - key: executable basefile, value: list of paths
        return: dictionary with key: direct dependency,
        value: dictionary with key transitive dependency element and value as package manager result for it
        """
        ld = LdDebugParser()
        ld_deb_dict = dict()
        ld_return_dict = dict()
        bdba_dict = dict()
        for executable, exec_list in executables.items():
            for each_ele in exec_list:
                str_ele = each_ele.strip(" ").strip('\"')
                ld_debug_dict = ld.run_ld_debug(str_ele, self.bld_dir)
                ld_deb_dict[str_ele] = ld_debug_dict
        for executable, components in ld_deb_dict.items():
            for dependency, dependency_of_dep in components.items():
                ld_return_dict[dependency], bdba_dict = self.check_pkg_for_ldds(list(dependency_of_dep), bdba_dict)
        return ld_return_dict, bdba_dict

    def check_components(self):
        """This function creates final dictionary joining all direct (from linker and header)
        and transitive dependencies from ldd and ld_debug
        """
        pkg_mgr_components = {**self.pkg_mgr_data['linker'], **self.pkg_mgr_data['header']}
        # logging.debug("pkg_manager_components now are {}".format(pkg_mgr_components))
        linker_components = self.ldd_linker
        ld_debug_components = self.lddebug_executable
        left_ovr_pairs_linker = set(linker_components.keys())
        left_ovr_pairs_ld_debug = set(ld_debug_components.keys())

        trans_dep_dict = defaultdict(set)
        for each_component, pkg_result in pkg_mgr_components.items():

            if each_component in linker_components:
                for dep_1, ldd_pkg_mgr_res in linker_components[each_component].items():
                    trans_dep_dict[pkg_result].add(ldd_pkg_mgr_res)
                left_ovr_pairs_linker.remove(each_component)

            if each_component in ld_debug_components:
                for dep_2, ld_pkg_mgr_res in ld_debug_components[each_component].items():
                    trans_dep_dict[pkg_result].add(ld_pkg_mgr_res)
                left_ovr_pairs_ld_debug.remove(each_component)

            if (each_component not in linker_components) and (each_component not in ld_debug_components):
                trans_dep_dict[pkg_result] = set()

        sub_linker_components = {key: linker_components[key] for key in left_ovr_pairs_linker if
                                 key in linker_components}
        sub_ld_debug_components = {key: ld_debug_components[key] for key in left_ovr_pairs_ld_debug if
                                   key in ld_debug_components}

        logging.debug("sub_linker_components is {}".format(sub_linker_components))
        logging.debug("sub_ld_debug_components is {}".format(sub_ld_debug_components))

        for dep, dep_trans in sub_linker_components.items():
            sha1, confidence, pkg_result_1, result = self.run_pkg_query([dep])
            if result:
                pkg_result_1 = self.join_components(sha1, confidence, pkg_result_1, self.direct_match_evidence)
                for dep_3, ld_pkg_mgr_r in dep_trans.items():
                    trans_dep_dict[pkg_result_1].add(ld_pkg_mgr_r)

        for dep, dep_trans in sub_ld_debug_components.items():
            sha1, confidence, pkg_result_1, result = self.run_pkg_query([dep])
            if result:
                pkg_result_1 = self.join_components(sha1, confidence, pkg_result_1, self.direct_match_evidence)
                for dep_3, ld_pkg_mgr_r in dep_trans.items():
                    trans_dep_dict[pkg_result_1].add(ld_pkg_mgr_r)
        return trans_dep_dict

    def create_csv(self, table_df):
        """This function creates raw_bdio.csv"""
        self.csv_fields_size()
        if table_df != []:
            if self.pkg_manager == 'rpm':
                df = pd.DataFrame(table_df,
                                  columns=["distro", "fullpath", "package-name", "package-version", "confidence",
                                           "size",
                                           "timestamp", "sha1", "matchType", "package-architecture", "epoch", "release",
                                           "trans-dep"])

            elif self.pkg_manager == 'dpkg':
                df = pd.DataFrame(table_df,
                                  columns=["distro", "fullpath", "package-name", "package-version", "confidence",
                                           "size",
                                           "timestamp", "sha1", "matchType", "package-architecture", "trans-dep"])
            elif self.pkg_manager == 'brew':
                df = pd.DataFrame(table_df,
                                  columns=["distro", "fullpath", "package-name", "package-version", "confidence",
                                           "size",
                                           "timestamp", "sha1", "matchType", "trans-dep"])
                df['package-architecture'] = 'none'
            df['separator'] = '/'
            df['package-type'] = self.pkg_manager
            df['type'] = 'distro-package'
            # csv_file_short = os.path.join(self.output_dir, "raw_bdio_short.csv")
            df[["size", "timestamp", "confidence"]] = df[["size", "timestamp", "confidence"]].apply(pd.to_numeric,
                                                                                                    errors='coerce')
            df[["size", "timestamp", "confidence"]] = df[["size", "timestamp", "confidence"]].fillna(value=0)
            df[["size", "timestamp", "confidence"]] = df[["size", "timestamp", "confidence"]].astype(int)
            df.fullpath = df.fullpath.apply(lambda x: tuple(x) if type(x) != str else tuple([x]))
            logging.debug("length of dataframe is {}".format(len(df)))
            df1 = df.groupby(
                ["package-name", "package-version", "package-architecture"]).agg({'trans-dep': sum}).reset_index()
            df1 = pd.merge(df1, df[
                ["distro", "fullpath", "package-name", "package-version", "package-architecture", "confidence", "size",
                 "timestamp", "sha1",
                 "matchType", "type", "package-type"]], on=["package-name", "package-version", "package-architecture"],
                           how='inner')
            df1['trans-dep'] = df['trans-dep'].apply(
                lambda x: [list(tdep_t) for tdep_t in set(tuple(tdep) for tdep in x)])
            duplicate_res = df1['trans-dep'].astype(str).duplicated()
            df1['trans-dep'] = df1['trans-dep'].where(~duplicate_res, "[]")
            df1['trans-dep'] = df1['trans-dep'].astype(str)
            df1['length'] = df1['trans-dep'].str.len()
            logging.debug("df1 columns are {}".format(df1.columns))
            df1.sort_values('length', ascending=False, inplace=True)
            df1.to_csv(self.csv_file_path,
                       columns=["distro", "fullpath", "package-name", "package-version", "package-architecture",
                                "confidence",
                                "size", "timestamp", "sha1", "matchType", "type", "package-type", "trans-dep"],
                       index=False)
            logging.info("raw_bdio.csv written to location {}".format(self.output_dir))
        else:
            df1 = pd.DataFrame(columns=["distro", "fullpath", "package-name", "package-version", "package-architecture",
                                        "confidence",
                                        "size", "timestamp", "sha1", "matchType", "type", "package-type", "trans-dep"])
            df1.to_csv(self.csv_file_path, index=False)
            logging.warning("Empty csv written..no package manager result found")

    def parse_components_info(self):
        """function to parse package manager result string into list of values"""
        table_df = []
        for component, trans_components_set in self.final_components_dict.items():
            col_df = self.parse_logic(component)
            trans_list = []
            for each_trans_comp in trans_components_set:
                trans_col_df = self.parse_logic(each_trans_comp, trans_status=True)
                trans_list.append(trans_col_df)
            trans_list.sort()
            uniq_trans_list = list(trans_list for trans_list, _ in itertools.groupby(trans_list))
            col_df.append(uniq_trans_list)
            table_df.append(col_df)
        self.create_csv(table_df)

    def parse_logic(self, component_str, trans_status=False):
        """ parsing logic for package manager result string
        param: string - package manager result
        return: list
        for eg: ['a/b/c.so','openssl','1.2.3',100,12436,2425364646,'faa816df1c632fcba5c4d525a180aa4c8b85f515',...]
        """
        component_info = []
        """ If os distribtion is mac/windows, setting it to fedora to get more external id matches """
        if self.os_dist == "Mac" or self.os_dist == "windows":
            self.os_dist = "fedora"
        component_info.append(self.os_dist)

        pattern_file = re.compile('(?<=file:)([^ ])*')
        fullpath_info = (self.bld_dir, re.search(pattern_file, component_str).group())
        pattern_pkg = re.compile('(?<=pkg:)([^ ])*')
        pkg_info = re.search(pattern_pkg, component_str).group()
        pattern_vers = re.compile('(?<=version:)([^ ])*')
        pkg_version_info = re.search(pattern_vers, component_str).group()
        pattern_conf = re.compile('(?<=confidence:)([^ ])*')
        try:
            confidence_info = int(re.search(pattern_conf, component_str).group())
        except ValueError:
            confidence_info = 0
        pattern_size = re.compile('(?<=size:)([^ ])*')
        try:
            size_info = int(re.search(pattern_size, component_str).group())
        except ValueError:
            size_info = 0
        pattern_time = re.compile('(?<=timestamp:)([^ ])*')
        try:
            time_info = int(re.search(pattern_time, component_str).group())
        except ValueError:
            time_info = 0
        pattern_sha1 = re.compile('(?<=sha1sum:)([^ ])*')
        sha1_info = re.search(pattern_sha1, component_str).group()

        pattern_matchtype = re.compile('(?<=matchType:)([^ ])*')
        matchtype_info = re.search(pattern_matchtype, component_str).group()

        if self.pkg_manager != 'brew':
            pattern_arch = re.compile('(?<=package-architecture:)([^ ])*')
            rx = re.compile('([()])')
            architecture_info = re.sub(rx, '', re.search(pattern_arch, component_str).group())
        if self.pkg_manager == 'rpm':
            pattern_epoch = re.compile('(?<=epoch:)([^ ])*')
            rx = re.compile('([()])')
            epoch_info = re.sub(rx, '', re.search(pattern_epoch, component_str).group())

            pattern_release = re.compile('(?<=release:)([^ ])*')
            release_info = re.sub(rx, '', re.search(pattern_release, component_str).group())

            pkg_version_info = epoch_info.replace('none:', '') + pkg_version_info + release_info.replace('-none', '')
            architecture_info = re.sub(rx, '', re.search(pattern_arch, component_str).group()).strip('/')

        component_info.append(fullpath_info)
        component_info.append(pkg_info)
        component_info.append(pkg_version_info)
        component_info.append(confidence_info)
        component_info.append(size_info)
        component_info.append(time_info)
        component_info.append(sha1_info)
        if trans_status:
            component_info.append('distro-package')
        component_info.append(matchtype_info)
        if self.pkg_manager != 'brew':
            component_info.append(architecture_info)
        if self.pkg_manager == 'rpm':
            component_info.append(epoch_info)
            component_info.append(release_info)
        return component_info

    def trans_key_value(self, string_in):
        """ function to append keys for transitive dependency list from csv for json
        param: string - package manager result for transitive dependency
        return: list of dictionaries
        eg: [{'distro':'ubuntu',..'package-name':'curl'..},
         {'distro':'ubuntu',..'package-name':'gcc'..}..]
        """
        dict_keys = ['distro', 'fullpath', 'package-name', 'package-version', 'confidence', 'size', 'timestamp', 'sha1',
                     'type',
                     'matchType', 'package-architecture']
        trans_dep_list = []
        list_in = ast.literal_eval(string_in)
        for each_list in list_in:
            if each_list != "[]":
                trans_dep_dict = dict(zip(dict_keys, each_list))
                trans_dep_list.append(trans_dep_dict)
        return trans_dep_list

    def csv_to_json(self):
        """This function converts raw_bdio.csv to raw_bdio.json"""
        data = {}
        try:
            with open(self.csv_file_path, encoding='utf-8') as csvf:
                csvReader = csv.DictReader(csvf)
                key1 = 'extended-objects'
                data['extended-objects'] = []
                for rows in csvReader:
                    new_row = dict()
                    for key, value in rows.items():
                        if key == 'trans-dep':
                            if value != "[]":
                                trans_dep_list = self.trans_key_value(value)
                                new_row[key] = trans_dep_list
                        elif key != 'fullpath':
                            try:
                                new_row[key] = int(value)
                            except ValueError:
                                new_row[key] = value
                        else:
                            new_row[key] = ast.literal_eval(value)
                    data[key1].append(new_row)
            try:
                with open(self.json_file_path, 'w', encoding='utf-8') as jsonf:
                    jsonf.write(json.dumps(data, indent=4))
                logging.info("raw_bdio.json written to location {}".format(self.output_dir))
            except OSError:
                logging.error("Couldn't write file {}".format(self.json_file_path))
        except FileNotFoundError:
            logging.error("raw_bdio.csv file not found - make sure csv file is written correctly")

    ## tar file
    def tar_files(self):
        """this function tars file for bdba"""
        tar_set_files = set()
        [[[tar_set_files.add(path) for path in paths] for paths in type.values()] for type in self.bdba_data.values()]
        logging.info("\nTarring all binary files\n")
        dir_path = os.path.join(self.output_dir, "bdba_ready.tar")
        if os.path.exists(dir_path):
            os.remove(dir_path)
        with tarfile.open(dir_path, 'w') as bdba_tar:
            for path in tqdm.tqdm(tar_set_files, total=len(tar_set_files)):
                if os.path.exists(path):
                    bdba_tar.add(path)
        logging.info("Files are placed in '{}' ".format(self.output_dir))
        logging.info("Number of files in bdba tar file are {}".format(len(tar_set_files)))

    def csv_fields_size(self):
        max_int = sys.maxsize
        while True:
            # decrease the maxInt value by factor 10  as long as the OverflowError occurs.
            try:
                csv.field_size_limit(max_int)
                break
            except OverflowError:
                max_int = int(max_int / 10)

    def get_dict_difference(self, json_dict, grep_dict):
        """ function to get files present in json and not in grep and vice versa
        param: json_dict: key with basefile, value as list of paths with that file from coverity json
               grep_dict: key with basefile, value as list of paths with that file from grep
        return: dict_json : files present in json not present in grep as dictionary with key as basefile and value as list of paths
                dict_grep : files present in grep not present in json as dictionary with key as basefile and value as list of paths
        """
        json_keys = set(json_dict.keys())
        # logging.debug("json keys are {}".format(json_keys))
        grep_keys = set(grep_dict.keys())
        # logging.debug("grep keys are {}".format(grep_keys))
        other_keys_in_json = json_keys - grep_keys
        other_keys_in_grep = grep_keys - json_keys
        dict_json = dict((key, json_dict[key]) for key in other_keys_in_json)
        dict_grep = dict((key, grep_dict[key]) for key in other_keys_in_grep)
        return dict_json, dict_grep

    def join_json_grep(self, json_dict, file_type):
        """ function to append resolved coverity json paths which were not found by grep to resolved_file_dict
        param: json_dict: key with basefile, value as list of paths with that file from coverity json
               file_type: 'header', 'linker' or 'executable'
        """
        # adding the values with common key
        for key in json_dict:
            if key in self.resolved_file_dict[file_type]:
                # logging.debug("self.resolved_file_dict[file_type][key] is {}".format(self.resolved_file_dict[file_type][key]))
                self.resolved_file_dict[file_type][key].union(json_dict[key])
            else:
                self.resolved_file_dict[file_type][key] = set(json_dict[key])
        # logging.debug("joined file_type {}".format(file_type))

    def skip_dynamic_files(self):
        self.linker_list = {each_file for each_file in self.linker_list if not (
                re.match(global_settings.so_pattern, each_file.strip()) or re.match(global_settings.dll_pattern,
                                                                                    each_file.strip()))}
        self.cov_linker_files_set = {each_file for each_file in self.cov_linker_files_set if not (
                re.match(global_settings.so_pattern, each_file.strip()) or re.match(global_settings.dll_pattern,
                                                                                    each_file.strip()))}

    def skip_pkg_mgr(self, basefiles_dict):
        bdba_dict = {}
        pkg_dict = {}
        for basefile, paths_list in basefiles_dict.items():
            bdba_dict[basefile] = paths_list
        return pkg_dict, bdba_dict

    def set_csv_matchtpye(self):
        """
        this function reads raw_bdio.csv and sets matchtype based on hub_version
        """
        if (PKG_MGR_MODE in self.run_modes or ALL_MODE in self.run_modes):  # and
            logging.info("Attempting to use offline files for package manager at location: {}".format(
                os.path.join(self.output_dir, 'raw_bdio.csv')))
            if os.path.exists(os.path.join(self.output_dir, 'raw_bdio.csv')):
                logging.info("found package manager csv file")
                raw_bdio_df = pd.read_csv(os.path.join(self.output_dir, 'raw_bdio.csv'))
                self.set_matchtype_per_hub_version()
                raw_bdio_df['matchType'] = self.direct_match_evidence
                raw_bdio_df['trans-dep'] = raw_bdio_df['trans-dep'].apply(
                    lambda x: x.replace('unknown', self.transitive_match_evidence))
                raw_bdio_df.to_csv(self.csv_file_path, index=False)
            else:
                logging.error(
                    "Unable to find previously generated offline files for package manager..please make sure use_offline_files and run_modes are set correctly")

    def set_matchtype_per_hub_version(self):
        """ this function gets hub version information and sets matchtype"""
        hub_version = self.hub_api.get_hub_version()
        logging.info("BLACK DUCK VERSION IS {}".format(hub_version))
        vers_result = hub_version.split(".")
        if int(vers_result[0]) >= 2021 and int(vers_result[1]) >= 6:
            self.direct_match_evidence = 'DIRECT_DEPENDENCY_BINARY'
            self.transitive_match_evidence = 'TRANSITIVE_DEPENDENCY_BINARY'
        else:
            self.direct_match_evidence = 'Binary'
            self.transitive_match_evidence = 'Binary'

    def run(self):
        run_mode_res = PKG_MGR_MODE in self.run_modes or ALL_MODE in self.run_modes
        if not self.use_offline_files:
            self.run_log_parser()

            if self.skip_dynamic:
                self.skip_dynamic_files()

            start_file_dict = datetime.now()
            self.file_dict['executable'] = self.get_files(self.executable_list)
            self.file_dict['linker'] = self.get_files(self.linker_list)
            self.file_dict['header'] = self.get_files(self.header_list)

            self.cov_file_dict['cov-executable'] = self.get_files(self.cov_executable_files_set)
            self.cov_file_dict['cov-linker'] = self.get_files(self.cov_linker_files_set)
            self.cov_file_dict['cov-header'] = self.get_files(self.cov_header_files_set)
            end_file_dict = datetime.now()
            logging.debug(
                "time taken to create dictionary from parsed and json files {}".format(end_file_dict - start_file_dict))

            logging.info("number of distinct executable files are {}".format(len(self.file_dict['executable'])))
            logging.info("number of distinct linker files are {}".format(len(self.file_dict['linker'])))
            logging.info("number of distinct header files are {}".format(len(self.file_dict['header'])))

            logging.info("coverity json - number of distinct executable files are {}".format(
                len(self.cov_file_dict['cov-executable'])))
            logging.info(
                "coverity json - number of distinct linker files are {}".format(len(self.cov_file_dict['cov-linker'])))
            logging.info(
                "coverity json - number of distinct header files are {}".format(len(self.cov_file_dict['cov-header'])))
            self.write_to_txt(self.file_dict, self.binary_files_path)
            self.unresol_set = set(
                list(self.file_dict['executable'].keys()) + list((self.file_dict['linker'].keys())) + list(
                    (self.file_dict['header'].keys())))
            logging.info("total number of unresolved files at beginning are : {}".format(len(self.unresol_set)))

            self.build_dir_dict = self.generate_build_dir_files()
            # logging.debug("number of distinct build dir files are {}".format(len(self.build_dir_dict)))

            start_resolv = datetime.now()
            self.resolved_file_dict['executable'], self.unresolved_file_dict['executable'] = self.resolve_file_paths(
                self.file_dict['executable'])
            self.resolved_file_dict['linker'], self.unresolved_file_dict['linker'] = self.resolve_file_paths(
                self.file_dict['linker'])
            self.resolved_file_dict['header'], self.unresolved_file_dict['header'] = self.resolve_file_paths(
                self.file_dict['header'])

            self.resolved_cov_file_dict['cov-executable'], self.unresolved_cov_file_dict[
                'cov-executable'] = self.resolve_file_paths(
                self.cov_file_dict['cov-executable'])
            self.resolved_cov_file_dict['cov-linker'], self.unresolved_cov_file_dict[
                'cov-linker'] = self.resolve_file_paths(
                self.cov_file_dict['cov-linker'])
            self.resolved_cov_file_dict['cov-header'], self.unresolved_cov_file_dict[
                'cov-header'] = self.resolve_file_paths(
                self.cov_file_dict['cov-header'])

            logging.debug("total resolved files: {} and total unresolved files : {} for executables".format(
                len(self.resolved_file_dict['executable']), len(self.unresolved_file_dict['executable'])))
            logging.debug("total resolved files: {} and total unresolved files : {} for linker".format(
                len(self.resolved_file_dict['linker']), len(self.unresolved_file_dict['linker'])))
            logging.debug("total resolved files: {} and total unresolved files : {} for header".format(
                len(self.resolved_file_dict['header']), len(self.unresolved_file_dict['header'])))

            logging.debug(
                "coverity json - total resolved files: {} and total unresolved files : {} for executables".format(
                    len(self.resolved_cov_file_dict['cov-executable']),
                    len(self.unresolved_cov_file_dict['cov-executable'])))
            logging.debug("coverity json - total resolved files: {} and total unresolved files : {} for linker".format(
                len(self.resolved_cov_file_dict['cov-linker']), len(self.unresolved_cov_file_dict['cov-linker'])))
            logging.debug("coverity json - total resolved files: {} and total unresolved files : {} for header".format(
                len(self.resolved_cov_file_dict['cov-header']), len(self.unresolved_cov_file_dict['cov-header'])))

            self.difference_dict = {}
            self.difference_dict['extra-exe-in-json'], self.difference_dict[
                'extra-exe-in-grep'] = self.get_dict_difference(self.resolved_cov_file_dict['cov-executable'],
                                                                self.resolved_file_dict['executable'])
            self.difference_dict['extra-linker-in-json'], self.difference_dict[
                'extra-linker-in-grep'] = self.get_dict_difference(self.resolved_cov_file_dict['cov-linker'],
                                                                   self.resolved_file_dict['linker'])
            self.difference_dict['extra-header-in-json'], self.difference_dict[
                'extra-header-in-grep'] = self.get_dict_difference(self.resolved_cov_file_dict['cov-header'],
                                                                   self.resolved_file_dict['header'])

            logging.info("total executables present in json not present in grep are : {}".format(
                len(self.difference_dict['extra-exe-in-json'])))
            logging.info("total linker present in json not present in grep are : {}".format(
                len(self.difference_dict['extra-linker-in-json'])))
            logging.info("total header present in json not present in grep are : {}".format(
                len(self.difference_dict['extra-header-in-json'])))
            logging.info("total executables present in grep not present in json are : {}".format(
                len(self.difference_dict['extra-exe-in-grep'])))
            logging.info("total linker present in grep not present in json are : {}".format(
                len(self.difference_dict['extra-linker-in-grep'])))
            logging.info("total header present in grep not present in json are : {}".format(
                len(self.difference_dict['extra-header-in-grep'])))

            self.write_to_txt(self.resolved_file_dict, self.resolved_files_path)
            self.write_to_txt(self.unresolved_file_dict, self.unresolved_files_path)
            self.write_to_txt(self.difference_dict, self.json_grep_files_list)
            self.write_to_txt(self.resolved_cov_file_dict, self.resolved_cov_files_path)
            self.write_to_txt(self.unresolved_cov_file_dict, self.unresolved_cov_files_path)

            # join here cov-json and grep
            self.join_json_grep(self.resolved_cov_file_dict['cov-linker'], 'linker')
            self.join_json_grep(self.resolved_cov_file_dict['cov-header'], 'header')
            self.join_json_grep(self.resolved_cov_file_dict['cov-executable'], 'executable')

            logging.debug("after join: total resolved files: {} and total unresolved files : {} for executables".format(
                len(self.resolved_file_dict['executable']), len(self.unresolved_file_dict['executable'])))
            logging.debug("after join: total resolved files: {} and total unresolved files : {} for linker".format(
                len(self.resolved_file_dict['linker']), len(self.unresolved_file_dict['linker'])))
            logging.debug("after join: total resolved files: {} and total unresolved files : {} for header".format(
                len(self.resolved_file_dict['header']), len(self.unresolved_file_dict['header'])))

            end_resolv = datetime.now()
            logging.debug(
                "time taken to resolve paths and join parse files with json files {}".format(end_resolv - start_resolv))

            # check hub version if online or take version input from user
        logging.debug("self.offline mode is {}".format(self.offline))
        logging.debug("self.use_offline_files is {}".format(self.use_offline_files))

        if self.offline:
            self.direct_match_evidence = 'unknown'
            self.transitive_match_evidence = 'unknown'
        else:
            self.set_matchtype_per_hub_version()

        if not self.use_offline_files:
            start_pkg = datetime.now()
            if run_mode_res:
                logging.debug("self.direct match evd is {}".format(self.direct_match_evidence))
                logging.debug("self.trans match evd is {}".format(self.transitive_match_evidence))

                self.pkg_mgr_data['linker'], self.bdba_data['linker'] = self.check_pkg_mgr(
                    self.resolved_file_dict['linker'])
                self.pkg_mgr_data['header'], self.bdba_data['header'] = self.check_pkg_mgr(
                    self.resolved_file_dict['header'])
            else:
                self.pkg_mgr_data['linker'], self.bdba_data['linker'] = self.skip_pkg_mgr(
                    self.resolved_file_dict['linker'])
                self.pkg_mgr_data['header'], self.bdba_data['header'] = self.skip_pkg_mgr(
                    self.resolved_file_dict['header'])
            end_pkg = datetime.now()
            logging.info("time taken to get package manager results {}".format(end_pkg - start_pkg))

            logging.debug("length of pkg_mgr_data for linker is {}".format(len(self.pkg_mgr_data['linker'])))
            logging.debug("length of bdba_data for linker is {}".format(len(self.bdba_data['linker'])))
            logging.debug("bdba_data for linker is {}".format(self.bdba_data['linker']))

            logging.debug("length of pkg_mgr_data for header is {}".format(len(self.pkg_mgr_data['header'])))
            logging.debug("length of bdba_data for header is {}".format(len(self.bdba_data['header'])))
            logging.debug("bdba_data for header is {}".format(self.bdba_data['header']))

            if self.can_perform_ldd() and not self.skip_transitives and not self.skip_dynamic and run_mode_res:
                start_ldd = datetime.now()
                self.ldd_linker, self.bdba_data['ldd_linker'] = self.check_ldd(self.pkg_mgr_data['linker'])
                logging.debug("ldd_linker is {}".format(self.ldd_linker))
                logging.debug("length of ldd_linker is {}".format(len(self.ldd_linker)))
                logging.debug("length of bdba_data for ldd_linker is {}".format(len(self.bdba_data['ldd_linker'])))
                logging.debug("bdba_data for ldd_linker is {}".format(self.bdba_data['ldd_linker']))

                self.lddebug_executable, self.bdba_data['lddebug_exe'] = self.check_ldd_deb(
                    self.resolved_file_dict['executable'])
                logging.debug("lddebug_executable is {}".format(self.lddebug_executable))
                logging.debug("length of lddebug_executable is {}".format(len(self.lddebug_executable)))
                logging.debug(
                    "length of bdba_data for lddebug_executable is {}".format(len(self.bdba_data['lddebug_exe'])))
                logging.debug("bdba_data for lddebug_executable is {}".format(self.bdba_data['lddebug_exe']))
                end_ldd = datetime.now()
                logging.info("time taken to get ldd results {}".format(end_ldd - start_ldd))

            if run_mode_res:
                self.final_components_dict = self.check_components()
                logging.debug("final_components_dict is  {}".format(self.final_components_dict))
                self.parse_components_info()  ## csv created in offline/online mode
                if not self.offline:
                    self.csv_to_json()
            if BDBA_MODE in self.run_modes or ALL_MODE in self.run_modes:
                self.tar_files()

        else:  ## if use_offline_files is set to True
            if run_mode_res:
                self.set_csv_matchtpye()
                self.csv_to_json()



"""
Copyright (c) 2021 Synopsys, Inc.
Use subject to the terms and conditions of the Synopsys End User Software License and Maintenance Agreement.
All rights reserved worldwide.
"""

import os
import re
import fnmatch
import ntpath
import logging
from blackduck_c_cpp.util import global_settings
import shlex


class LogParser:
    """
    Parses a build log output file from coverity build command
    """

    def __init__(self):
        self.lib_path = set()
        self.dash_l_path = set()
        self.all_paths = set()
        self.abs_paths = set()
        self.rpath_binary_paths = set()
        self.pkgconf_file_paths = dict()
        self.executable_list = set()

    def resolve_path(self, path):
        """
            Function to resolve a path that contains "../" segments
            For example /lib/a/b/../c -> /lib/a/c
        """
        new_path = []
        for component in path.split('/'):
            if len(component) > 0:
                if component == '..':
                    new_path = new_path[0:len(new_path) - 1]
                else:
                    new_path = new_path + [component]
        return '/' + '/'.join(new_path)

    def find_file(self, pattern, path):
        """
            Fuzzy find files from a root path.
        """
        result = []
        for root, dirs, files in os.walk(path):
            result.extend((os.path.join(root, x) for x in fnmatch.filter(files, pattern)))
        return result

    def path_leaf(self, path):
        """
            Windows-safe filename finder.
        """
        head, tail = ntpath.split(path)
        return tail or ntpath.basename(head)

    def parse_windows(self, log_file):
        """
            parsing build-log.txt for windows
        """
        try:
            with open(log_file, 'r') as text_file:
                for line in text_file:
                    match_list = [re.match(global_settings.cplusplus_pattern, line),
                                  re.match(global_settings.gplusplus_pattern, line),
                                  re.match(global_settings.cc_pattern, line),
                                  re.match(global_settings.gcc_pattern, line)]

                    non_match_list = [re.match(global_settings.compl_non_mtch_ptrn_1, line),
                                      re.match(global_settings.compl_non_mtch_ptrn_2, line)]

                    linker_list = [re.match(global_settings.ld_pattern, line),
                                   re.match(global_settings.mvsc_link_pattern, line),
                                   re.match(global_settings.mvsc_cl_pattern, line)]

                    if any(linker_list) or (any(match_list) and not any(non_match_list)):
                        line_components = shlex.split(line, posix=False)
                        line_components = set(
                            map(lambda x: x.replace("/Fo", "") if x.startswith("/Fo") else x, line_components))

                        for exec_line_idx, exec_line in enumerate(line_components):
                            if exec_line.lower().startswith('/libpath:'):
                                self.lib_path.add(self.resolve_path(exec_line.lower().replace('/libpath:', '')))
                            elif exec_line.lower().startswith('/out:'):
                                self.all_paths.add(self.resolve_path(exec_line.lower().replace('/out:', '')))
                            elif re.match(global_settings.dll_pattern, exec_line.strip()) or re.match(
                                    global_settings.o_pattern, exec_line.strip()) \
                                    or re.match(global_settings.obj_pattern, exec_line.strip()):
                                self.all_paths.add(self.resolve_path(exec_line.strip('\n"')))

        except FileNotFoundError:
            logging.error("build-log.txt not found from cov-build..make sure cov-build is successful")

        basefile_list = set(map(lambda path: os.path.basename(path).strip(), self.all_paths))
        if len(basefile_list) == 0:
            logging.warning("No linker files found")
        logging.info("No of distinct components are {}".format(len(basefile_list)))
        self.all_paths.update(basefile_list)
        self.all_paths = sorted(self.all_paths, key=self.key_fn)
        return self.all_paths

    def parse_other(self, log_file):
        """
            parsing build-log.txt for linux,mac,fedora,redhat..
        """
        try:
            with open(log_file, 'r') as text_file:
                for line in text_file:
                    match_list = [re.match(global_settings.cplusplus_pattern, line),
                                  re.match(global_settings.gplusplus_pattern, line),
                                  re.match(global_settings.cc_pattern, line),
                                  re.match(global_settings.gcc_pattern, line)]

                    non_match_list = [re.match(global_settings.compl_non_mtch_ptrn_1, line),
                                      re.match(global_settings.compl_non_mtch_ptrn_2, line)]

                    linker_list = [re.match(global_settings.ld_pattern, line)]

                    if any(linker_list) or (any(match_list) and not any(non_match_list)):
                        line_components = re.split(',|\s+', line)
                        line_components = set(map(lambda x: x.strip('"').strip("'"), line_components))
                        in_rpath = False  # are we inside of an rpath section?
                        rpath_path_set = None  # path(s) the rpath defines.
                        rpath_library_set = None  # the libraries that will be linked to said paths
                        for exec_idx, exec_line in enumerate(line_components):
                            if exec_line == "-rpath":
                                in_rpath = True
                                rpath_path_set = set()
                                rpath_library_set = set()
                            elif in_rpath:  # if we're in an "-rpath", now capture the paths, and the binaries.
                                if (":" in exec_line and all(
                                        [os.path.exists(x) for x in exec_line.split(":")])) or os.path.exists(
                                    exec_line):
                                    rpath_path_set = set(exec_line.split(":"))
                                elif not exec_line.startswith("-"):  # when '-' shows up, we're out of the rpath
                                    for x in rpath_path_set:
                                        if os.path.exists(self.resolve_path(os.path.join(x, exec_line))):
                                            self.rpath_binary_paths.add(self.resolve_path(os.path.join(x, exec_line)))
                                        else:
                                            self.rpath_binary_paths.add(exec_line)
                                    # self.rpath_binary_paths = self.rpath_binary_paths.union(set([self.resolve_path(os.path.join(x, lc))
                                    #                                                    for x in rpath_path_set if os.path.exists(self.resolve_path(os.path.join(x, lc)))]))
                                    rpath_library_set.add(exec_line)
                                else:
                                    in_rpath = False
                                # TODO: In future, when adding pkg config
                                """
                                    for path in rpath_path_set:
                                        pkgconf_files = self.find_file("*.pc", path)
                                        if len(pkgconf_files) > 0:
                                            for pc_path in pkgconf_files:
                                                f_name = self.path_leaf(pc_path).split('.')[0]
                                                if f_name not in self.pkgconf_file_paths:
                                                    self.pkgconf_file_paths[f_name] = pc_path
                                """
                            if not in_rpath:
                                if exec_line.startswith('-L') and '/home' not in exec_line:
                                    self.lib_path.add(self.resolve_path(exec_line.replace('-L', '')))
                                if exec_line.startswith('-l'):
                                    self.dash_l_path.add(exec_line.strip('\n '))
                                elif re.match(global_settings.so_pattern, exec_line.strip()) or re.match(
                                        global_settings.dll_pattern, exec_line.strip()) \
                                        or re.match(global_settings.a_pattern, exec_line.strip()) or re.match(
                                    global_settings.lib_pattern, exec_line.strip()) \
                                        or re.match(global_settings.o_pattern, exec_line.strip()):
                                    self.all_paths.add(self.resolve_path(exec_line.strip('\n"')))
            # logging.debug("rpath binaries are : {}".format(self.rpath_binary_paths))
            self.all_paths.union(self.rpath_binary_paths)
            # logging.debug("all paths from build log are : {}".format(self.all_paths))

        except FileNotFoundError:
            logging.error("build-log.txt not found from cov-build..make sure cov-build is successful")

        for base_path in self.lib_path:
            for lib in self.dash_l_path:
                libname = lib.replace('-l', 'lib')
                x_so = "{}/{}.so".format(base_path, libname)
                x_a = "{}/{}.a".format(base_path, libname)
                self.all_paths.add(x_a)
                self.all_paths.add(x_so)
        basefile_list = set(map(lambda path: os.path.basename(path).strip(), self.all_paths))
        if len(basefile_list) == 0:
            logging.warning("No linker files found")
        logging.info("No of distinct components are {}".format(len(basefile_list)))
        self.all_paths.update(basefile_list)
        self.all_paths = sorted(self.all_paths, key=self.key_fn)
        return self.all_paths

    def parse_build_log(self, log_file, os_dist):
        """
        Parses on build-log.txt file and returns files called in linker invocation
        """

        if os_dist == 'windows':
            self.parse_windows(log_file)
        else:
            self.parse_other(log_file)
        return self.all_paths

    def key_fn(self, key):
        """
        Sort function to sort the paths in the order of .so, .a and .o files
        """
        if re.match(global_settings.so_pattern, key):
            return (0, key)
        elif re.match(global_settings.a_pattern, key):
            return (1, key)
        else:
            return (2, key)

    def get_exectuable(self, log_file):
        """ Parses on build-log.txt file to get output executable
               """
        try:
            with open(log_file, 'r') as text_file:
                for line in text_file:
                    match_list = [re.match(r".*EXECUTING: .*", line), re.match(r".* -o .*", line)]
                    if all(match_list):
                        pattern = re.compile('(?<=-o )([^ ])*')
                        executable_mid = re.search(pattern, line).group()
                        if not re.match(global_settings.o_pattern, executable_mid) and not re.match(
                                global_settings.cpp_pattern, executable_mid) \
                                and not re.match(global_settings.so_pattern, executable_mid) and not re.match(
                            global_settings.hpp_pattern, executable_mid) \
                                and not re.match(global_settings.a_pattern, executable_mid) and not re.match(
                            global_settings.h_pattern, executable_mid) \
                                and not re.match(global_settings.c_pattern, executable_mid) and not re.match(
                            global_settings.dll_pattern, executable_mid):
                            if not re.match(r'.*\.s$', executable_mid):
                                self.executable_list.add(re.search(pattern, line).group())
        except FileNotFoundError:
            logging.error("build-log.txt not found from cov-build..make sure cov-build is successful")
        return self.executable_list



"""
Copyright (c) 2021 Synopsys, Inc.
Use subject to the terms and conditions of the Synopsys End User Software License and Maintenance Agreement.
All rights reserved worldwide.
"""


def parse_go_pseudoversion(version):
    # Pseudoversion
    if version.startswith("v") and "." in version:
        tail = version.split(".")[-1]
        tailparts = tail.split("-")
        # Check if matches pseudoversion format
        if (
            len(tailparts) > 1
            and len(tailparts[-2]) == 14
            and len(tailparts[-1]) == 12
            and tailparts[-2].isdigit()
        ):
            return tailparts[-2], tailparts[-1]
    raise ValueError("Not pseudo-version")


"""
Copyright (c) 2021 Synopsys, Inc.
Use subject to the terms and conditions of the Synopsys End User Software License and Maintenance Agreement.
All rights reserved worldwide.
"""

from collections import defaultdict
from blackduck_c_cpp.util import util
import os
import subprocess
import re
import pandas as pd
from blackduck_c_cpp.util import global_settings


class LdDebugParser:

    def __init__(self):
        pass

    def initialize(self, nodes):
        self.graph = defaultdict(list)
        self.n = nodes

    def add_edge(self, u, v):
        self.graph[u].append(v)

    def run_ld_debug(self, executable, bld_dir):
        ld_debug_dict = defaultdict(set)
        if os.path.exists(executable):
            out = subprocess.getstatusoutput("LD_DEBUG=files ldd " + "{}".format(executable) + " 2> /dev/null")
        elif os.path.exists(os.path.join(bld_dir, executable)):
            out = subprocess.getstatusoutput(
                "LD_DEBUG=files ldd " + "{}".format(os.path.join(bld_dir, executable)) + " 2> /dev/null")
        else:
            return ld_debug_dict

        status, ld_result = util.value_getstatusoutput(out)
        output_path_set = defaultdict(list)

        if status and ('no path found matching pattern' not in ld_result):
            parsed_results = re.findall(r'^\s*(\S+)\s+=>(.*) \(.*\)$', ld_result, re.MULTILINE)
            for fname, path in parsed_results:
                """ storing filename: fullpath eg: a.so: g/f/d/a.so"""
                output_path_set[fname.strip()] = path.strip()
            pattern_transitive = re.compile('(?<=file=)([^ ])*')
            pattern_direct = re.compile('(?<=needed by )([^ ])*')
            self.initialize(list(output_path_set.values()))
            row_df = []
            for each_line in ld_result.split("\n"):
                col_df = []
                res_transitive = re.search(pattern_transitive, each_line)
                res_direct = re.search(pattern_direct, each_line)
                if res_transitive and res_direct:
                    res_direct_path = res_direct.group().strip()
                    res_transitive_path = res_transitive.group().strip()
                    res_direct_basename = os.path.basename(res_direct_path)
                    res_transitive_basename = os.path.basename(res_transitive_path)

                    if res_direct_basename not in output_path_set:
                        output_path_set[res_direct_basename] = res_direct_path

                    self.add_edge(output_path_set[res_direct_basename], output_path_set[res_transitive_basename])
                    col_df.append(output_path_set[res_direct_basename])
                    col_df.append(output_path_set[res_transitive_basename])
                    row_df.append(col_df)
            dep_df = pd.DataFrame(row_df, columns=['dep', 'dep_of_dep'])
            dep_df = dep_df[dep_df.astype(str)['dep'] != '[]']
            dep_df = dep_df[dep_df.astype(str)['dep_of_dep'] != '[]']
            ld_debug_dict = defaultdict(set)
            for dep, dep_dep in zip(list(dep_df['dep']), list(dep_df['dep_of_dep'])):
                ld_debug_dict[dep].add(dep_dep)
            dep_of_dep_list = list(ld_debug_dict.values())
            dep_list = list(ld_debug_dict.keys())
            keys_to_del = []
            """if dep is present in list of transitive dependencies,
            add everything to dependency and del transitive dependency key"""
            for dep, dep_of_dep in ld_debug_dict.items():
                if re.match(global_settings.so_pattern, dep):
                    dep_key = [dep_list[idx] for idx, trans_dep in enumerate(dep_of_dep_list) if dep in trans_dep]
                    if re.match(global_settings.so_pattern, dep_key[0]):
                        ld_debug_dict[dep_key[0]].update(ld_debug_dict[dep])
                        keys_to_del.append(dep)
            for each_key in keys_to_del:
                del ld_debug_dict[each_key]
            return ld_debug_dict
        else:
            return ld_debug_dict



"""
Copyright (c) 2021 Synopsys, Inc.
Use subject to the terms and conditions of the Synopsys End User Software License and Maintenance Agreement.
All rights reserved worldwide.
"""

import sys
import os
import logging
import json
import shlex
from datetime import datetime
from blackduck_c_cpp.util import util
from blackduck_c_cpp.util.c_arg_parser import C_Parser
from blackduck_c_cpp.pkg_manager.pkg_manager_detector import PkgDetector
from blackduck_c_cpp.pkg_manager.pkg_manager_bom import PkgManagerBom
from blackduck_c_cpp.bdba.upload_bdba import BDBAApi
from blackduck_c_cpp.sig_scanner.emit_wrapper import EmitWrapper
from blackduck_c_cpp.bdio.bdio2_python_transformer import BDIO2Transformer
from blackduck_c_cpp.sig_scanner.run_sig_scanner import SigScanner
from blackduck_c_cpp.util.hub_api import HubAPI
from blackduck_c_cpp.coverity_json.cov_json_parse import CovJsonParser
from blackduck_c_cpp.util.cov_install import CoverityInstall
import glob
import platform
import re
from os.path import expanduser
import pkg_resources
import stat
import subprocess

"""
This class will run coverity build capture for the supplied build command.
It assumes that any setup and configuration for the build has already taken place 
prior to calling this code.
"""


class RunBuildCapture(object):

    def __init__(self, my_args, log_path, blackduck_output_dir):
        self.my_args = my_args
        self.start_time = datetime.now()
        self.blackduck_c_cpp_version = self.get_package_version()
        self.home = expanduser("~")
        self.bld_dir = self.my_args.args.build_dir
        self.bld_cmd = self.my_args.args.build_cmd
        self.hub_project_name = self.my_args.args.project_name
        self.hub_project_vers = shlex.quote(self.my_args.args.project_version)
        self.cov_dir = self.my_args.args.cov_output_dir if self.my_args.args.cov_output_dir else os.path.join(self.home,
                                                                                                              ".synopsys",
                                                                                                              "blackduck-c-cpp",
                                                                                                              "output",
                                                                                                              self.hub_project_name)
        self.blackduck_output_dir = blackduck_output_dir
        self.skip_build = my_args.args.skip_build
        self.bd_url = shlex.quote(self.my_args.args.bd_url.strip('/'))
        self.codelocation_name = shlex.quote(
            self.my_args.args.codelocation_name) if self.my_args.args.codelocation_name else "{}/{}".format(
            self.hub_project_name, self.hub_project_vers)
        self.additional_sig_scan_args = self.my_args.args.additional_sig_scan_args
        self.expand_sig_files = self.my_args.args.expand_sig_files
        self.additional_coverity_params = self.my_args.args.additional_coverity_params
        self.api_token = shlex.quote(self.my_args.args.api_token) if self.my_args.args.api_token else os.environ.get(
            'BLACKDUCK_API_TOKEN')
        self.insecure = True if self.my_args.args.insecure else False
        self.force = True if self.my_args.args.force else False
        self.skip_transitives = self.my_args.args.skip_transitives
        self.skip_includes = self.my_args.args.skip_includes
        self.skip_dynamic = self.my_args.args.skip_dynamic
        self.use_offline_files = self.my_args.args.use_offline_files
        self.offline_mode = self.my_args.args.offline
        self.scan_cli_dir = self.my_args.args.scan_cli_dir if self.my_args.args.scan_cli_dir is not None else self.blackduck_output_dir
        self.cov_configure_args = my_args.args.cov_configure_args
        self.run_modes = list(map(lambda mode: mode.lower(), map(lambda s: re.sub("[\"\']", "", s),
                                                                 map(str.strip, self.my_args.args.modes.split(",")))))
        logging.debug("run modes is {}".format(self.run_modes))
        self.PKG_MGR_MODE = 'pkg_mgr'
        self.SIG_MODE = 'sig'
        self.BDBA_MODE = 'bdba'
        self.ALL_MODE = 'all'
        self.hub_api = HubAPI(self.bd_url, self.api_token, self.insecure)
        self.log_path = log_path
        self.platform_name = platform.system().lower()
        self.cov_home = self.set_coverity_root()
        self.cov_bin = os.path.join(self.cov_home, "bin")
        self.build_log = os.path.join(self.cov_dir, "build-log.txt")
        self.binary_files_list = os.path.join(self.blackduck_output_dir, "all_binary_paths.txt")
        self.json_grep_files_list = os.path.join(self.blackduck_output_dir, "json_grep_files_diff.txt")
        self.unresolved_files_list = os.path.join(self.blackduck_output_dir, "unresolved_file_paths.txt")
        self.resolved_files_list = os.path.join(self.blackduck_output_dir, "resolved_file_paths.txt")
        self.resolved_cov_files_list = os.path.join(self.blackduck_output_dir, "resolved_cov_file_paths.txt")
        self.unresolved_cov_files_list = os.path.join(self.blackduck_output_dir, "unresolved_cov_file_paths.txt")
        self.cov_json_file_path = os.path.join(self.blackduck_output_dir, "cov_emit_links.json")

    def __enter__(self):
        if not self.offline_mode:
            self.hub_api.create_or_verify_project_version_exists(self.hub_project_name, self.hub_project_vers)
        if not os.path.exists(self.cov_dir):
            os.makedirs(self.cov_dir)

        print()
        logging.info("RUNNING BLACKDUCK-C-CPP VERSION {}".format(self.blackduck_c_cpp_version))
        logging.info("Output log is written to file {}".format(os.path.join(self.log_path)))
        logging.info("coverity output files will be in {}".format(self.cov_dir))
        logging.debug("Arguments passed into blackduck-c-cpp are {}".format(self.my_args.args))

        offline_yaml_location = os.path.join(self.blackduck_output_dir, "offline_config.yaml")
        if self.offline_mode:
            logging.debug("Storing config file to location : {}".format(offline_yaml_location))
            with open(offline_yaml_location, 'w') as f:
                json.dump(self.my_args.args.__dict__, f, indent=2)
        if self.use_offline_files:
            logging.warning(
                "Please make sure skip_* parameters are kept exactly same as in offline mode from offline configuration file {}".format(
                    offline_yaml_location))

        if self.use_offline_files and (self.offline_mode or not self.skip_build):
            util.error_and_exit("Cannot set use_offline_files=True and offline=True/skip_build=False")

        if self.insecure:
            logging.warning("SSL verification has been disabled with the --insecure flag")

        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_type:
            print(f'exc_type: {exc_type}')
            print(f'exc_value: {exc_value}')
            print(f'exc_traceback: {exc_traceback}')

    @staticmethod
    def handle_uncaught_exp(type, value, traceback):
        logging.error("Logging an uncaught exception", exc_info=(type, value, traceback))

    def get_phase_status_dict(self):
        """
        Read the log file and check for errors in each section
        :return: a dict of each phase and its status (SUCCESS or ERROR)
        """
        phase_dict = {}
        phase_pattern = r'Phase: (.*) '
        error_pattern = r'ERROR: (.*)$'

        try:
            with open(self.log_path, 'r') as f:
                contents = f.read()
        except:
            logging.error("Could not open log file to generate a status report - verify directory permissions")
            return phase_dict

        phases = re.findall(phase_pattern, contents, flags=re.MULTILINE)
        log_sections = re.split(phase_pattern, contents)

        for p in phases:
            if 'Blackuck-c-cpp Status Report' not in p:
                content_index = log_sections.index(p) + 1
                error_matches = re.findall(error_pattern, log_sections[content_index], flags=re.MULTILINE)
                if not error_matches:
                    phase_dict[p] = 'SUCCESS'
                if error_matches:
                    phase_dict[p] = 'ERROR - SEE LOGS AT {} FOR DETAILS'.format(self.log_path)
        return phase_dict

    @staticmethod
    def get_package_version():
        """
        Get the version of blackduck-c-cpp that's being run
        Will only work if the blackduck-c-cpp module has been installed - will return unknown if running from source
        code that's been added to PYTHONPATH
        :return: the blackduck-c-cpp version
        """
        try:
            version_c_cpp_tool = pkg_resources.require("blackduck-c-cpp")[0].version
            return version_c_cpp_tool
        except Exception as e:
            logging.warning("unable to get version of blackduck-c-cpp")
            return 'UNKNOWN'

    @staticmethod
    def log_phase(phase):
        """
        Call at the beginning of each phase to log the current process being run
        :param: the current phase
        """
        print()  # line break between phases for readability
        logging.info("******************************   Phase: {} ***********************************".format(phase))

    def get_cov_version(self, cov_home):
        cmd = os.path.join(cov_home, "bin", "cov-build") + ' --ident'
        out = subprocess.getstatusoutput(cmd)
        status, cov_version = util.value_getstatusoutput(out)
        if status:
            return cov_version
        else:
            return 'UNKNOWN'

    def set_coverity_root(self):
        self.log_phase("Get coverity root information")
        if self.my_args.args.coverity_root and os.path.exists(self.my_args.args.coverity_root):
            logging.debug("Using coverity package given by user")
            cov_home = self.my_args.args.coverity_root
        else:
            cov_home = self.coverity_download()
        logging.info("coverity files are at {}".format(cov_home))
        logging.info("COVERITY VERSION IS {}".format(self.get_cov_version(cov_home)))
        return cov_home

    def coverity_download(self):
        """
        download coverity if cov_root is not given by user
        :return: the files to be tarred by the sig scanner
        """
        start_download = datetime.now()
        logging.debug("Downloading Coverity as coverity_root is not given by user")
        cov_home = os.path.join(expanduser("~"), ".synopsys", "blackduck-c-cpp", "cov-build-capture")
        if not os.path.exists(cov_home):
            os.makedirs(cov_home)
        CoverityInstall(cov_home, self.hub_api, self.platform_name, self.force, self.blackduck_c_cpp_version)
        end_download = datetime.now()
        logging.debug('Time taken to download coverity files: {}'.format(end_download - start_download))
        return cov_home

    def run_build(self, cov_emit_link_path, cov_build_path):
        """
        Run cov-build to build the c/cpp target project
        """
        self.log_phase("Starting Build")
        sys.excepthook = self.handle_uncaught_exp
        start_build = datetime.now()
        cov_conf = os.path.join(self.blackduck_output_dir, "conf")

        # to write build command to script
        if not os.path.exists(self.bld_cmd):
            if 'windows' in self.platform_name:
                sh_path = os.path.join(self.cov_dir, 'cov-build.bat')
            else:
                sh_path = os.path.join(self.cov_dir, 'cov-build.sh')
            with open(sh_path, 'w') as rsh:
                rsh.write("{}".format(self.bld_cmd))
            st = os.stat(sh_path)
            os.chmod(sh_path, st.st_mode | stat.S_IEXEC)
        else:
            sh_path = self.bld_cmd

        if not os.path.exists(cov_conf):
            os.makedirs(cov_conf)

        logging.debug("PLATFORM NAME IS {}".format(self.platform_name))
        if 'windows' in self.platform_name:
            conf_key = " & "
        else:
            conf_key = " \n "

        conf_cmd = util.conf_cmd("--clang", self.cov_bin, cov_conf)
        conf_cmd += conf_key + util.conf_cmd("--gcc", self.cov_bin, cov_conf)
        conf_cmd += conf_key + util.conf_cmd("--msvc", self.cov_bin, cov_conf)
        conf_cmd += conf_key + util.conf_cmd("--template --compiler cc --comptype gcc", self.cov_bin, cov_conf)

        conf_cmd += conf_key + util.conf_cmd("--template --compiler gcc --comptype gcc", self.cov_bin, cov_conf)
        conf_cmd += conf_key + util.conf_cmd("--template --compiler *gcc* --comptype gcc", self.cov_bin, cov_conf)

        conf_cmd += conf_key + util.conf_cmd("--template --compiler *c++* --comptype gcc", self.cov_bin, cov_conf)
        conf_cmd += conf_key + util.conf_cmd("--template --compiler *g++* --comptype gcc", self.cov_bin, cov_conf)

        conf_cmd += conf_key + util.conf_cmd("--template --compiler *g++* --comptype g++", self.cov_bin, cov_conf)

        conf_cmd += conf_key + util.conf_cmd("--template --compiler ld --comptype gcc", self.cov_bin, cov_conf)
        conf_cmd += conf_key + util.conf_cmd("--template --compiler *-ld --comptype gcc", self.cov_bin, cov_conf)
        conf_cmd += conf_key + util.conf_cmd("--template --compiler ^ld-* --comptype gcc", self.cov_bin, cov_conf)

        conf_cmd += conf_key + util.conf_cmd("--template --compiler ld --comptype ld", self.cov_bin, cov_conf)
        conf_cmd += conf_key + util.conf_cmd("--template --compiler *-ld --comptype ld", self.cov_bin, cov_conf)
        conf_cmd += conf_key + util.conf_cmd("--template --compiler ^ld-* --comptype ld", self.cov_bin, cov_conf)

        conf_cmd += conf_key + util.conf_cmd("--template --compiler ccache --comptype prefix", self.cov_bin, cov_conf)
        for compiler, comptype in self.cov_configure_args.items():
            conf_cmd += conf_key + util.conf_cmd("--template --compiler {} --comptype {}".format(compiler, comptype),
                                                 self.cov_bin, cov_conf)

        if self.additional_coverity_params:
            cov_add_cmd = ' {}'.format(self.additional_coverity_params)
        else:
            cov_add_cmd = ''

        if os.path.exists(cov_emit_link_path):
            self.bld_cmd = '"{}" --emit-link-units -c "{}" --dir "{}"{} "{}"'.format(cov_build_path,
                                                                                     os.path.join(cov_conf, "bld.xml"),
                                                                                     self.cov_dir, cov_add_cmd, sh_path)
        else:
            self.bld_cmd = '"{}" -c "{}" --dir "{}"{} "{}"'.format(cov_build_path, os.path.join(cov_conf, "bld.xml"),
                                                                   self.cov_dir, cov_add_cmd, sh_path)

        full_cmd = conf_cmd + conf_key + self.bld_cmd
        logging.info("Build Command: {}".format(self.bld_cmd))
        ret_code = util.run_cmd(full_cmd, self.bld_dir)
        end_build = datetime.now()
        logging.info('Time taken to run build: {}'.format(end_build - start_build))
        if ret_code != 0:
            util.error_and_exit(
                "Build not successful, please make sure to do successful build - return code {}".format(ret_code))

    def run_cov_emit(self):
        """
        Run cov manage emit
        :return: the files to be tarred by the sig scanner
        :return: the header files on which to run the package manager scan
        """
        self.log_phase("Running Cov Manage Emit")
        start_emit = datetime.now()
        if not self.use_offline_files:
            blackduck_emit_wrapper = EmitWrapper(self.cov_home, self.cov_dir, self.blackduck_output_dir,
                                                 self.platform_name)
            emit_wrapper_output_sig = blackduck_emit_wrapper.cov_emit_output_sig
            emit_cov_header_files = blackduck_emit_wrapper.cov_header_files
        else:
            emit_wrapper_output_sig = emit_cov_header_files = None
            if os.path.exists(os.path.join(self.blackduck_output_dir, 'cov_emit_output_files')):
                logging.info("Attempting to use offline files for cov-manage-emit at location: {}".format(
                    os.path.join(self.blackduck_output_dir, 'cov_emit_output_files')))
            else:
                logging.error(
                    "Unable to find previously generated offline files for cov-manage-emit, set use_offline_files to false to generate new ones.")
        end_emit = datetime.now()
        logging.info('Time taken to get emit files: {}'.format(end_emit - start_emit))

        return emit_wrapper_output_sig, emit_cov_header_files

    def run_cov_emit_link(self, cov_manage_emit_path):
        """
        Run cov-emit-link
        """
        self.log_phase("Generating Cov-emit-link json")
        start_emit_link = datetime.now()
        header_files_set = linker_files_set = executable_files_set = set()
        if not self.use_offline_files:
            full_cmd = '"{}" --dir "{}" list-capture-invocations > "{}"'.format(cov_manage_emit_path,
                                                                                self.cov_dir,
                                                                                os.path.join(self.blackduck_output_dir,
                                                                                             "cov_emit_links.json"))
            ret_code = util.run_cmd(full_cmd, self.cov_dir)
            if ret_code == 0:
                cov_parser = CovJsonParser(self.cov_json_file_path)
                header_files_set, linker_files_set, executable_files_set = cov_parser.run()
            else:
                logging.error("coverity json not created successfully- return code {}".format(ret_code))

        end_emit_link = datetime.now()
        logging.info('Time taken to get emit link json files if present: {}'.format(end_emit_link - start_emit_link))
        return header_files_set, linker_files_set, executable_files_set

    def run_get_package_manager_bom(self, emit_cov_header_files, header_files_set, linker_files_set,
                                    executable_files_set):
        """
        Get the package manager BOM, to be written to a BDIO file later in run_bdio_from_package_manager()
        :return: the package manager BOM from which to creaete the BDIO file
        :return: the current OS
        """
        self.log_phase("Getting package manager BOM")
        # if not use_offline_files:
        start_pkg_mgr = datetime.now()
        pkg_manager, os_dist = PkgDetector().get_platform()
        # skips all header files from package manager and bdba matching if skip_includes is True
        if self.skip_includes:
            emit_cov_header_files = []
            header_files_set = []
        if self.use_offline_files:
            emit_cov_header_files = []
            header_files_set = []
        if pkg_manager and os_dist:
            pkg_mgr_bom = PkgManagerBom(pkg_manager,
                                        self.build_log,
                                        self.cov_home,
                                        self.blackduck_output_dir,
                                        os_dist,
                                        self.bld_dir,
                                        self.unresolved_files_list,
                                        self.resolved_files_list,
                                        self.binary_files_list,
                                        self.json_grep_files_list,
                                        self.skip_build,
                                        self.skip_transitives,
                                        self.skip_dynamic,
                                        self.run_modes,
                                        emit_cov_header_files, self.my_args.args.debug, header_files_set,
                                        linker_files_set, executable_files_set, self.resolved_cov_files_list,
                                        self.unresolved_cov_files_list, self.hub_api,
                                        self.offline_mode, self.use_offline_files)
            pkg_mgr_bom.run()
        else:
            pkg_mgr_bom = None
        end_pkg_mgr = datetime.now()
        logging.info('Time taken to get package manager results: {}'.format(end_pkg_mgr - start_pkg_mgr))
        return pkg_mgr_bom, os_dist

    def run_bdba(self):
        """
        Tar files for BDBA and run the BDBA scan
        """
        self.log_phase("BDBA")
        start_bdba = datetime.now()
        if not self.use_offline_files:
            logging.info(
                "The bdba tar file to scan will be written to {}".format(
                    os.path.join(self.blackduck_output_dir, "bdba_ready.tar")))
        else:
            if os.path.exists(os.path.join(self.blackduck_output_dir, "bdba_ready.tar")):
                logging.info("Attempting to use offline files for BDBA at location: {}".format(
                    os.path.join(self.blackduck_output_dir, 'bdba_ready.tar')))
            else:
                logging.error(
                    "Unable to find previously generated offline files for BDBA..set use_offline_files to false to generate new ones.")
        if not self.offline_mode:
            BDBAApi(self.hub_api).upload_binary(self.hub_project_name,
                                                self.hub_project_vers,
                                                self.codelocation_name,
                                                os.path.join(self.blackduck_output_dir, "bdba_ready.tar"))
        end_bdba = datetime.now()
        logging.info('Time taken to get bdba results: {}'.format(end_bdba - start_bdba))

    def run_sig_scan(self, emit_wrapper_output_sig, os_dist, pkg_mgr_bom_bdba_data=None):
        """
        Download the sig scanner if appropriate, tar files to be scanned, and run the sig scanner
        :param emit_wrapper_output_sig: the files to be tarred and scanned by the sig scanner
        :param os_dist: the current OS
        :param pkg_mgr_bom_bdba_data: a dict of header files to additionally scan
        """
        self.log_phase("BD Signature Scan")
        start_sig = datetime.now()
        sig_scanner = SigScanner(self.cov_home,
                                 self.hub_api,
                                 pkg_mgr_bom_bdba_data,
                                 self.hub_project_name,
                                 self.hub_project_vers,
                                 self.codelocation_name,
                                 emit_wrapper_output_sig,
                                 self.blackduck_output_dir,
                                 self.offline_mode,
                                 self.scan_cli_dir,
                                 self.use_offline_files,
                                 os_dist,
                                 self.additional_sig_scan_args,
                                 self.expand_sig_files,
                                 self.my_args.args.scan_interval,
                                 self.my_args.args.json_splitter_limit,
                                 self.my_args.args.disable_json_splitter,
                                 self.my_args.args.skip_includes)
        sig_scanner.run()
        end_sig = datetime.now()
        logging.info('Time taken to get sig results: {}'.format(end_sig - start_sig))

    def run_bdio_from_package_manager(self):
        """
        Use the results from run_get_package_manager_bom() to create a BDIO file and upload it to Black Duck
        """
        self.log_phase("BDIO from our Package Manager")
        start_bdio = datetime.now()
        bdio2_python_wrapper = BDIO2Transformer(self.blackduck_output_dir, self.hub_api)
        json_file_path = os.path.join(self.blackduck_output_dir, "raw_bdio.json")
        if not os.path.exists(json_file_path):
            logging.warning("The raw BDIO file was not generated. There will not be BDIO results.")
            return
        with open(json_file_path, 'r') as f:
            raw_bdio_json = json.load(f)
        cov_results = [raw_bdio_json]  # Load results from input raw_bdio json
        scan_size = os.stat(json_file_path).st_size  # size of jsonld bdio2 file
        skip_bdio2 = False
        try:
            scanned_path = cov_results[0]['extended-objects'][0]['fullpath'][0]
            if not scanned_path:  # get the scanned path
                scanned_path = "/"
        except IndexError:
            logging.debug(
                "no components present in json {}".format(os.path.join(self.blackduck_output_dir, "raw_bdio.json")))
            skip_bdio2 = True
        if not skip_bdio2:
            result_bdio2_str, code_loc_name = bdio2_python_wrapper.generate_bdio2(self.codelocation_name,
                                                                                  self.hub_project_name,
                                                                                  self.hub_project_vers,
                                                                                  cov_results, scanned_path, scan_size)
            # Rest Remains same above code then code be swapped out once bogdan release his library changes
            bdio2_output_filename = bdio2_python_wrapper.write_bdio2_document(result_bdio2_str)
            logging.info("Created bdio2 file: %s", bdio2_output_filename)
            logging.info("Uploading bdio2 to Black Duck: %s", bdio2_output_filename)
            response = self.hub_api.hub.upload_scan(bdio2_output_filename)
            if response.ok:
                print(bdio2_output_filename, "uploaded successfully !!")
            else:
                logging.error("Error uploading {} bdio2 file -- (Response({}): {})".format(bdio2_output_filename,
                                                                                           response.status_code,
                                                                                           response.text))
        end_bdio = datetime.now()
        logging.info('Time taken to get bdio results: {}'.format(end_bdio - start_bdio))

    def final_status(self):
        """
        At the end of the blackduck-c-cpp run, get the status of each phase of the blackduck-c-cpp run and log
        whether it was successful or if it contained errors
        """
        self.log_phase("Blackuck-c-cpp Status Report")
        logging.info("Blackuck-c-cpp run is complete")
        phase_dict = self.get_phase_status_dict()

        phases = phase_dict.keys()
        for item in phases:
            logging.info("Phase {} status: {}".format(item, phase_dict[item]))
        logging.info('Time taken to run: {}'.format(datetime.now() - self.start_time))

    def run_build_capture(self):
        """
        Run each phase of the blackduck-c-cpp process
        """

        cov_emit_link_path = ""

        try:
            cov_build_path = glob.glob(os.path.join(self.cov_bin, 'cov-build*'))[0]
            cov_manage_emit_path = glob.glob(os.path.join(self.cov_bin, 'cov-manage-emit*'))[0]
        except IndexError:
            util.error_and_exit(
                "cov-build or cov-manage-emit not found in bin directory at location: {}".format(self.cov_bin))

        try:
            cov_emit_link_path = glob.glob(os.path.join(self.cov_bin, 'cov-emit-link*'))[0]
        except IndexError:
            pass

        # COV BUILD
        if not self.skip_build:
            self.run_build(cov_emit_link_path, cov_build_path)

        # COV EMIT
        emit_wrapper_output_sig, emit_cov_header_files = self.run_cov_emit()

        # COV EMIT LINK
        if os.path.exists(cov_emit_link_path):
            header_files_set, linker_files_set, executable_files_set = self.run_cov_emit_link(cov_manage_emit_path)
        else:
            logging.info("Not performing coverity json parsing due to absence of cov-emit-link in bin directory")
            header_files_set = linker_files_set = executable_files_set = set()

        # PACKAGE MANAGER BOM
        pkg_mgr_bom, os_dist = self.run_get_package_manager_bom(emit_cov_header_files, header_files_set,
                                                                linker_files_set,
                                                                executable_files_set)

        # BDBA
        if self.BDBA_MODE in self.run_modes or self.ALL_MODE in self.run_modes:
            self.run_bdba()

        # SIG SCAN
        if self.SIG_MODE in self.run_modes or self.ALL_MODE in self.run_modes:
            pkg_mgr_bom_bdba_data = None
            if pkg_mgr_bom:
                pkg_mgr_bom_bdba_data = pkg_mgr_bom.bdba_data
            self.run_sig_scan(emit_wrapper_output_sig, os_dist, pkg_mgr_bom_bdba_data)

        # BDIO
        if not self.offline_mode and (self.PKG_MGR_MODE in self.run_modes or self.ALL_MODE in self.run_modes):
            self.run_bdio_from_package_manager()

        # STATUS
        self.final_status()


def setup_logging(log_path, verbose):
    """
    Set up logging for the module
    """
    format_str = "%(filename)s:%(funcName)s:%(lineno)d:%(asctime)s:%(levelname)s: %(message)s"
    logging.basicConfig(format=format_str, level='DEBUG', filename=log_path, filemode='w')
    console = logging.StreamHandler()
    console.setLevel(level=logging.DEBUG if verbose else logging.INFO)
    console.setFormatter(logging.Formatter(format_str))
    logging.getLogger('').addHandler(console)


def run():
    """
    Entry point
    """
    my_args = C_Parser()
    blackduck_output_dir = my_args.args.output_dir if my_args.args.output_dir else os.path.join(expanduser("~"),
                                                                                                ".synopsys",
                                                                                                "blackduck-c-cpp",
                                                                                                "output",
                                                                                                my_args.args.project_name)

    if not os.path.exists(blackduck_output_dir):
        os.makedirs(blackduck_output_dir)

    log_path = os.path.join(blackduck_output_dir, 'blackduck_c_cpp.log')
    setup_logging(log_path, my_args.args.verbose)

    with RunBuildCapture(my_args, log_path, blackduck_output_dir) as RBC:
        RBC.run_build_capture()


if __name__ == '__main__':
    run()



"""
Copyright (c) 2021 Synopsys, Inc.
Use subject to the terms and conditions of the Synopsys End User Software License and Maintenance Agreement.
All rights reserved worldwide.
"""
from google.cloud import storage
from google.oauth2 import service_account
import hashlib
import shutil
import logging
import json
from blackduck_c_cpp.util import util
import os
import stat
import tempfile
from blackduck_c_cpp.util import global_settings


class CoverityInstall:
    """Class to download latest coverity package for different OS """

    def __init__(self, cov_home, hub_api, platform, force, blackduck_c_cpp_version):
        hub_api.authenticate()
        self.hub_api = hub_api
        self.hub = hub_api.hub
        self.cov_home = cov_home
        self.platform = platform
        if self.platform == global_settings.darwin_pltfrm:
            self.platform = global_settings.macos_pltfrm
            platform = global_settings.macos_pltfrm
        self.force = force
        self.cov_zippath = 'cov-latest' + '-' + platform + '.zip'
        self.cov_filepath = os.path.join(self.cov_home, self.cov_zippath)
        self.tool_token_response = None
        self.tool_token_json = None
        self.gcp_filepath = None
        self.blackduck_c_cpp_version = blackduck_c_cpp_version
        try:
            self.run()
        except Exception as e:
            logging.error("Exception occurred: {}".format(e))

    def hub_api_tool_token(self):
        """
        get token from hub tool-token api
        """
        data = json.dumps({"clientName": "blackduck-c-cpp", "clientVersion": self.blackduck_c_cpp_version})
        self.tool_token_response = self.hub.execute_post('{}/api/tool-token'.format(self.hub.config['baseurl']), data)
        if not self.tool_token_response.ok:
            logging.error(
                "Problem connecting to /api/tool-token -- (Response({}): {})".format(
                    self.tool_token_response.status_code,
                    self.tool_token_response.text))
        self.tool_token_json = self.tool_token_response.json()

    def find_any_coverity_version(self):
        """check if any version of coverity is present"""
        if os.path.exists(self.cov_filepath):
            logging.debug(
                "Using older coverity version already present on the system from {}".format(
                    self.cov_filepath))
        else:
            util.error_and_exit("No coverity version is present on the system. Please contact Synopsys Support")

    def gcp_download(self):
        """
        Download mini coverity package from gcp if credentials are correct
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            tempfile_path = os.path.join(tmpdir, self.cov_zippath)
            try:
                storage_credentials = service_account.Credentials.from_service_account_info(
                    self.tool_token_json['jsonWebToken'])
                client = storage.Client(credentials=storage_credentials)
                gcp_path_split = self.gcp_filepath.split("/")
                object_name = gcp_path_split[-1]
                bucket_name = gcp_path_split[-2]
                bucket = client.bucket(bucket_name)
                blob = bucket.blob(object_name)
                logging.info("Downloading coverity latest version from gcp to {}".format(tempfile_path))
                blob.download_to_filename(tempfile_path)
            except Exception as e:
                logging.warning("unable to connect to gcp - {}".format(e))
                self.find_any_coverity_version()
                return
            if not self.gcp_checksum == hashlib.sha256(open(tempfile_path, 'rb').read()).hexdigest():
                if not self.force:
                    logging.error(
                        "There's a mismatch between api tool token and gcp checksums, If you want to continue with previous version of coverity, please set force to True in yaml file and retry")
                    util.error_and_exit(
                        "PLEASE CONTACT SYNOPSYS SUPPORT IMMEDIATELY: Mismatch between GCP Bucket and Tool Token Sha2.")
                else:
                    self.find_any_coverity_version()
            else:
                if os.path.exists(self.cov_filepath):
                    os.remove(self.cov_filepath)
                dest = shutil.move(tempfile_path, self.cov_filepath)
                logging.debug("moving coverity files from {} to {}".format(tempfile_path, dest))
                self.unzip_downloaded()

    def unzip_downloaded(self):
        """
        unzip coverity package
        """
        logging.debug("Unzipping coverity latest version zip to location {}".format(self.cov_home))
        shutil.unpack_archive(self.cov_filepath, self.cov_home)

    def verify_checksum_match(self):
        """
        download coverity only if checksum doesn't match
        """
        self.gcp_checksum, self.gcp_filepath = self.tool_token_json['checksum'][self.platform].split(",")
        logging.debug("checksum of latest coverity file in gcp is {}".format(self.gcp_checksum))
        if os.path.exists(self.cov_filepath):
            sha2_filepath = hashlib.sha256(open(self.cov_filepath, 'rb').read()).hexdigest()
            if self.gcp_checksum == sha2_filepath:
                logging.info("Skipping downloading latest version of coverity as it already exists")
                return True
            else:
                logging.info("Downloading latest coverity package as it is not present on disk")
                return False
        else:
            return False

    def set_permissions(self):
        """ set executable permissions for coverity files"""
        logging.debug("setting executable permissions for coverity files")
        st = os.stat(self.cov_home)
        os.chmod(self.cov_home, st.st_mode | stat.S_IEXEC)
        for root, dirs, files in os.walk(self.cov_home):
            for d in dirs:
                st = os.stat(os.path.join(root, d))
                os.chmod(os.path.join(root, d), st.st_mode | stat.S_IEXEC)
            for f in files:
                st = os.stat(os.path.join(root, f))
                os.chmod(os.path.join(root, f), st.st_mode | stat.S_IEXEC)

    def run(self):
        """
        run coverity install
        """
        hub_version = self.hub_api.get_hub_version()
        vers_result = hub_version.split(".")
        if int(vers_result[0]) >= 2021 and int(vers_result[1]) >= 10:
            self.hub_api_tool_token()
            match_result = self.verify_checksum_match()
            if not match_result:
                self.gcp_download()
            else:
                logging.debug("Latest version is already present on the system at {}".format(self.cov_home))
            self.set_permissions()
        else:
            util.error_and_exit(
                "Automatic update of coverity not supported in Black Duck version {}. please provide coverity_root in yaml file".format(
                    hub_version))



"""
Copyright (c) 2021 Synopsys, Inc.
Use subject to the terms and conditions of the Synopsys End User Software License and Maintenance Agreement.
All rights reserved worldwide.
"""

from blackduck.HubRestApi import HubInstance
from blackduck_c_cpp.util import util
from requests.exceptions import ConnectionError


class HubAPI:

    def __init__(self, bd_url, api_token, insecure):
        self.bd_url = bd_url
        self.api_token = api_token
        self.insecure = insecure
        self.hub = None

    def authenticate(self):
        try:
            self.hub = HubInstance(self.bd_url, api_token=self.api_token, insecure=self.insecure,
                                   write_config_flag=False)
        except KeyError:
            util.error_and_exit("Make sure you have the right API key and --insecure flag is set correctly")
        except ConnectionError:
            util.error_and_exit(
                "ConnectionError occurred. Make sure you have the correct url and insecure flag is set correctly")

    def create_or_verify_project_version_exists(self, project_name, version_name, phase='DEVELOPMENT'):
        self.authenticate()
        project = self.hub.get_project_by_name(project_name)
        if project:
            version = self.hub.get_version_by_name(project, version_name)
            if version:
                return
            if not version:
                response = self.hub.create_project_version(project, version_name, parameters={'phase': phase})
        else:
            response = self.hub.create_project(project_name, version_name, parameters={'version_phase': phase})

        if not response.ok:
            util.error_and_exit(
                "Error creating project version -- (Response({}): {})".format(response.status_code, response.text))

    def get_hub_version(self):
        hub_version_dict = {}
        try:
            self.authenticate()
            hub_version_dict = self.hub.version_info
        except:  # in case we don't get version info
            hub_version_dict['version'] = '0.0.0'
        return hub_version_dict['version']



"""
Copyright (c) 2021 Synopsys, Inc.
Use subject to the terms and conditions of the Synopsys End User Software License and Maintenance Agreement.
All rights reserved worldwide.

Run coverity/bin/cov-manage-emit and capture output
"""

import logging
import os
import re
import shutil
from shutil import copy
from blackduck_c_cpp.util import util
import sys
from blackduck_c_cpp.util import global_settings
import glob


class EmitWrapper:

    def __init__(self, cov_base_path, cov_output_path, blackduck_output_dir, platform_name):

        self.cov_base_path = cov_base_path
        self.cov_output_path = cov_output_path
        self.blackduck_output_dir = blackduck_output_dir
        self.platform_name = platform_name
        self.cov_emit_output_files_path = os.path.join(self.blackduck_output_dir, 'cov_emit_output_files')
        self.cov_header_files = {}
        self.cov_emit_output_sig = {}

        logging.info("Files output by cov-emit will be copied to {}".format(self.cov_emit_output_files_path))

        try:
            self.run_cov_emit()
        except Exception as e:
            logging.error("Exception occurred: {}".format(e))
            logging.error("emit directory is empty - please make sure cov-build and cov-emit ran successfully ")

    def get_latest_emit_dir(self):
        """
        get the most recently modified directory in the emit directory in the cov output directory
        """
        try:
            search_dir = os.path.join(self.cov_output_path, 'emit')
            directories = [os.path.join(search_dir, x) for x in os.listdir(search_dir) if
                           os.path.isdir(os.path.join(search_dir, x))]
            directories.sort(key=lambda x: os.path.getmtime(x), reverse=True)
            if directories:
                return directories[0]
        except FileNotFoundError:
            util.error_and_exit("emit directory not present - please verify if coverity build is successful and cov-build files are present in {}".format(
                    self.cov_output_path))
        return None

    def run_cov_emit(self):
        """
        run cov-manage-emit
        """
        logging.info("Running cov-emit...")
        try:
            cov_emit_path = glob.glob(os.path.join(self.cov_base_path, 'bin', 'cov-manage-emit*'))[0]
        except IndexError:
            util.error_and_exit("cov-manage-emit not present in location: {}".format(os.path.join(self.cov_base_path, 'bin')))
        if 'windows' in self.platform_name:
            str_cmd = 'findstr /v /c:"Translation unit:"'
        else:
            str_cmd = 'grep -v "Translation unit:"'

        tu_pattern_command = '"{}" --dir "{}" --tu-pattern \"all()\" print-source-files | {} | awk {{\'print $3\'}}'.format(
            cov_emit_path, self.cov_output_path, str_cmd)

        list_command = '"{}" --dir "{}" list | {} | awk {{\"print $3\"}}'.format(
            cov_emit_path, self.cov_output_path, str_cmd)
        logging.debug("tu_pattern command is {}".format(tu_pattern_command))
        logging.debug("list_command is {}".format(list_command))
        cov_emit_result = util.run_cmd_emit(tu_pattern_command) + util.run_cmd_emit(list_command)

        self.cov_emit_output_sig = {x for x in cov_emit_result if x != ':' and os.path.exists(x)}
        self.cov_header_files = {x for x in cov_emit_result if x != ':' and (
                re.match(global_settings.hpp_pattern, x.strip()) or re.match(global_settings.h_pattern,
                                                                             x.strip())) and os.path.exists(x)}

        logging.info("Total cov emit output files: {}".format(len(self.cov_emit_output_sig)))
        logging.info("Total header files: {}".format(len(self.cov_header_files)))

        if os.path.exists(self.cov_emit_output_files_path):
            shutil.rmtree(self.cov_emit_output_files_path)
        os.makedirs(self.cov_emit_output_files_path)

        for f in self.cov_emit_output_sig:
            try:
                copy(f, self.cov_emit_output_files_path)
            except PermissionError:
                pass
            except IsADirectoryError:
                pass
                # logging.debug("File copy failed because file already exists in target directory {}".format(f))



"""
Copyright (c) 2021 Synopsys, Inc.
Use subject to the terms and conditions of the Synopsys End User Software License and Maintenance Agreement.
All rights reserved worldwide.
"""

import platform
import logging
import subprocess
from blackduck_c_cpp.util import util


class PkgDetector:
    """
    Detects Operating system, its distribution and package manager
    """

    def __init__(self):
        # default values to skip package manager steps
        self.package_manager = 'not_found'
        self.os_distribution = 'not_found'

    """ 
    Here are some notes to enhance this method for expanded usefulness:
    /etc/os-release       // ubuntu, redhat, centos or debian (etc... - look inside)
    /etc/SuSE-release     // SUSE 
    /etc/redhat-release   // redhat
    /etc/redhat_version   // redhat
    /etc/system-release   // centos, redhat
    /etc/slackware-release // Slackware
    /etc/fedora-release   // fedora
    /etc/alpine-release   // alpine
    /etc/debian_release   // old debian
    /etc/debian-version   // debian or ubuntu
    /etc/lsb-release      // debian or ubuntu
    /etc/mandrake-release // Mandrake
    /etc/yellowdog-release // Yellow dog
    /etc/gentoo-release    // Gentoo
    /etc/sun-release       // Sun JDS
    /etc/release          // Solaris/Sparc
    """

    def get_platform(self):
        """
        Returns package manager and os distribution
        """
        try:
            platform_name = platform.system()
            logging.info("Platform name is: {}".format(platform_name))
            if 'linux' in platform_name.lower():
                if util.run_cmd("rpm -qa --quiet", suppress_output=True) == 0:
                    self.package_manager = "rpm"
                    self.os_distribution = "fedora"  # default distribution if /etc/os-release file not found
                elif util.run_cmd("dpkg --help", suppress_output=True) == 0:
                    self.package_manager = "dpkg"
                    self.os_distribution = "ubuntu"  # default distribution if /etc/os-release file not found
                (exit_code, os_dist) = subprocess.getstatusoutput("cat /etc/os-release | grep HOME_URL")
                if exit_code == 0:
                    if 'ubuntu' in os_dist.lower():
                        self.os_distribution = "ubuntu"
                    elif 'centos' in os_dist.lower():
                        self.os_distribution = "centos"
                    elif 'fedora' in os_dist.lower():
                        self.os_distribution = "fedora"
                    elif 'opensuse' in os_dist.lower():
                        self.os_distribution = "opensuse"
                    elif 'alpine' in os_dist.lower():
                        self.os_distribution = "alpine"
                    else:
                        pass
            elif 'darwin' in platform_name.lower():
                if util.run_cmd("brew --help") == 0:
                    self.package_manager = 'brew'
                    self.os_distribution = 'Mac'
            elif 'windows' in platform_name.lower():
                self.package_manager = 'not_found'
                self.os_distribution = 'windows'
            logging.info("Operating system distribution is {}".format(self.os_distribution))
            logging.info("Package manager is {}".format(self.package_manager))
            return self.package_manager, self.os_distribution
        except:
            logging.warning("Operating system distribution is not found")
            logging.warning("Package manager is not found")
            return None, None



import re

so_pattern = re.compile(r".*\.so$|.*\.so\..+$")
a_pattern = re.compile(r".*\.a$|.*\.a\..+$")
o_pattern = re.compile(r".*\.o$|.*\.o\..+$")
obj_pattern = re.compile(r".*\.obj$|.*\.obj\..+$")
cpp_pattern = re.compile(r".*\.cpp$|.*\.cpp\..+$")
c_pattern = re.compile(r".*\.c$|.*\.c\..+$")
hpp_pattern = re.compile(r".*\.hpp$|.*\.hpp\..+$")
h_pattern = re.compile(r".*\.h$|.*\.h\..+$")
dll_pattern = re.compile(r".*\.dll$|.*\.dll\..+$")
lib_pattern = re.compile(r".*\.lib|.*\.lib\..+$")

cplusplus_pattern = re.compile(r".*EXECUTING: .*\Sc\+\+ ")
gplusplus_pattern = re.compile(r".*EXECUTING: .*\Sg\+\+ ")
cc_pattern = re.compile(r".*EXECUTING: .*\Scc ")
gcc_pattern = re.compile(r".*EXECUTING: .*\Sgcc ")
ld_pattern = re.compile(r".*EXECUTING: .*\Sld ")
mvsc_link_pattern = re.compile(r".*EXECUTING: LINK", re.IGNORECASE)
mvsc_cl_pattern = re.compile(r".*EXECUTING: CL", re.IGNORECASE)
compl_non_mtch_ptrn_1 = re.compile(r".* -c .*")
compl_non_mtch_ptrn_2 = re.compile(r".* -E .*")

snippet_match = 'snippet'
darwin_pltfrm = 'darwin'
macos_pltfrm = 'macos'



"""
Copyright (c) 2021 Synopsys, Inc.
Use subject to the terms and conditions of the Synopsys End User Software License and Maintenance Agreement.
All rights reserved worldwide.
"""

import json
from uuid import uuid4, uuid5, NAMESPACE_URL
from datetime import datetime
from typing import List, Dict, Tuple, Optional
from . import bdio_paths
from ..custom_types import JsonDict, ExtendedFullPath
from ..version_king import parse_go_pseudoversion


def count_components(bdio: JsonDict) -> int:
    return len(get_components(bdio))


def get_components(bdio: JsonDict) -> List[JsonDict]:
    components = get_bdio_entries_by_type(bdio, "Component")
    if components:
        return components
    return get_bdio_entries_by_type(bdio, "https://blackducksoftware.github.io/bdio#Component")


def get_files(bdio: JsonDict) -> List[JsonDict]:
    files = get_bdio_entries_by_type(bdio, "File")
    if files:
        return files
    return get_bdio_entries_by_type(bdio, "https://blackducksoftware.github.io/bdio#File")


def get_bdio_entries_by_type(bdio: JsonDict, type_id: str) -> List[JsonDict]:
    result: List[JsonDict] = []  # for the dear stupid mypy who cannot infer this but just nags
    if "@graph" not in bdio:
        return result
    for entry in bdio["@graph"]:
        type_value = entry.get("@type")
        if isinstance(type_value, list) and type_value[0] == type_id:
            result.append(entry)
        elif type_value == type_id:
            result.append(entry)
    return result


def get_all_extended_fileentries(scanner_results: List[JsonDict]) -> List[ExtendedFullPath]:
    result = []
    for component_info in scanner_results:
        for extended_obj in component_info["extended-objects"]:
            extended_fullpath = extended_obj.get("extended-fullpath")
            if not extended_fullpath:
                continue  # some old scans do not have those
            result.append(extended_fullpath)
    return result


def get_all_fileentries_bdio(bdio_content: JsonDict) -> List[ExtendedFullPath]:
    extended_fullpaths: List[ExtendedFullPath] = []
    for file_entry_bdio in get_files(bdio_content):
        if "path" in file_entry_bdio:
            file_path_bdio = file_entry_bdio["path"]
        elif "https://blackducksoftware.github.io/bdio#hasPath" in file_entry_bdio:
            file_path_bdio = file_entry_bdio["https://blackducksoftware.github.io/bdio#hasPath"][0][
                "@value"
            ]
        else:
            continue
        extended_fullpath = bdio_paths.bdio_path_uri_to_fullpath(file_path_bdio)
        extended_fullpaths.append(extended_fullpath)
    return extended_fullpaths


def split_version_from_external_id(forge: str, identifier: str) -> Tuple[str, str]:
    name_version_cutoff = 1
    if forge == "maven":
        # Maven entries have group and artifact parts separated with separator
        name_version_cutoff = 2
    separator = BDIOEntries.forge_to_separator_char.get(forge, "/")
    parts = identifier.split(separator)
    return separator.join(parts[:name_version_cutoff]), separator.join(parts[name_version_cutoff:])


def dump_bdio_to_file(bdio: JsonDict, filepath: str) -> None:
    with open(filepath, "w") as fd:
        json.dump(bdio, fd, indent=4, sort_keys=True, ensure_ascii=False)


def mangle_go_version(version: Optional[str]) -> Optional[str]:
    """
    BD KB Go component versions are in pre-go.mod format.
    Mangle go.mod -style versions to match KB.
    """
    # Non go.mod dependency
    if not version:
        return version

    if version.endswith("+incompatible"):
        version = version.replace("+incompatible", "")

    try:
        _, gittag = parse_go_pseudoversion(version)
        version = gittag
    except ValueError:
        pass
    return version


def _format_datetime(date: datetime) -> str:
    # BD datetime parser needs an offset at the end
    return date.isoformat() + "Z"


#    return date.strftime("%Y-%m-%dT%H:%M:%SZ")
# is fixed in py3 by using the timezone class from datetime and giving that as an argument
# when retrieving the time. Then isoformat() will add it to the print


class BDIOEntries:
    """
    Low level factory methods to create JSON dicts according to the BDIO syntax
    for various entities and datatypes used in BDBA.
    For combining the produced entries see and use the BDIOBuilder.

    See also https://github.com/blackducksoftware/bdio/tree/master/bdio2/src/main/java/com/blackducksoftware/bdio2/model
    for standard implementation.
    """

    forge_to_separator_char = {
        "alpine": "/",
        "anaconda": "/",
        "android_sdk": "/",
        "apache_software": "/",
        "arch_linux": "/",
        "automotive_linux": "/",
        "bower": "/",
        "centos": "/",
        "clearlinux": "/",
        "codeplex": "/",
        "codeplex_group": "/",
        "cpan": "/",
        "cran": "/",
        "crates": "/",
        "dart": "/",
        "debian": "/",
        "eclipse": "/",
        "fedora": "/",
        "freedesktop_org": "/",
        "github_gist": "/",
        "googlecode": "/",
        "hackage": "/",
        "hex": "/",
        "java_net": "/",
        "kde_org": "/",
        "launchpad": "/",
        "mongodb": "/",
        "npmjs": "/",
        "nuget": "/",
        "openembedded": "/",
        "opensuse": "/",
        "oracle_linux": "/",
        "pear": "/",
        "pypi": "/",
        "redhat": "/",
        "ros": "/",
        "rubyforge": "/",
        "rubygems": "/",
        "runtime": "/",
        "sourceforge": "/",
        "sourceforge_jp": "/",
        "tianocore": "/",
        "ubuntu": "/",
        "yocto": "/",
        "alt_linux": ":",
        "android": ":",
        "bitbucket": ":",
        "cocoapods": ":",
        "cpe": ":",
        "efisbot": ":",
        "gitcafe": ":",
        "github": ":",
        "gitlab": ":",
        "gitorious": ":",
        "gnu": ":",
        "golang": ":",
        "maven": ":",
        "openjdk": ":",
        "packagist": ":",
        "protecode_sc": ":",
        "long_tail": "#",
        "conan": "$",
    }

    @staticmethod
    def component(forge: str, identifier: str, fallback_forge: bool = False) -> JsonDict:
        component = {
            "@type": "Component",
            "@id": f"urn:uuid:{uuid4()}",
            "identifier": identifier,
        }
        if fallback_forge:
            # the @ denotes a class of forges, the Hub looks into alternative forges aka
            # fallback forges if it cannot find stuff in this one
            component["namespace"] = f"@{forge}"
        else:
            component["namespace"] = f"{forge}"

        return component

    @staticmethod
    def fileentry(
        extended_fullpath: ExtendedFullPath,
        timestamp: int,
        sha1: str,
        filesize: Optional[int] = None,
        annotation: Optional[JsonDict] = None,
        original_scanned_filename: Optional[str] = None,
    ) -> JsonDict:
        """
        Describe a file with its path and various other attributes.
        """
        if original_scanned_filename:
            # first file has a name in the scan project that might have been changed during the
            # upload to some temporary random file name, so patch it back here
            extended_fullpath[0] = extended_fullpath[0].copy()
            extended_fullpath[0]["path"] = original_scanned_filename

        fileentry_bdio: JsonDict = {  # required typing due to later int value added
            "@type": "File",
            "@id": f"urn:uuid:{uuid4()}",
            "path": bdio_paths.fullpath_to_bdio_path_uri(extended_fullpath),
            # choices for us below are: "regular", "regular/binary",
            # "directory/archive", "regular/text"
            # TODO: omit this for now as bdio linter chokes on wrong types
            # and currently we do not have any logic below to omit the right type, e.g.directory
            # "fileSystemType": "regular/binary",
            "fingerprint": f"sha1:{sha1}",
            "lastModifiedDateTime": _format_datetime(datetime.utcfromtimestamp(timestamp)),
        }
        # some older results might not have the filesize - introduced around 2019
        if filesize:
            fileentry_bdio["byteCount"] = filesize
        if annotation:
            fileentry_bdio["description"] = annotation["@id"]
        return fileentry_bdio

    @staticmethod
    def fileentry_project_base(scanned_path: str, filesize: Optional[int] = None) -> JsonDict:
        extended_fullpath = [{"type": "file", "path": scanned_path}]
        bdio_path = bdio_paths.fullpath_to_bdio_path_uri(extended_fullpath)
        # if no component findings were in the toplevel file then we will not have a cached
        # fileentry_from_extended_fullpath for it and need to create one here with not all file properties known
        fileentry_bdio: JsonDict = {
            "@type": "File",
            "@id": f"urn:uuid:{uuid4()}",
            "path": bdio_path,
        }
        if filesize:
            fileentry_bdio["byteCount"] = filesize
        # TODO: would need to find out if it was a file or directory that we scanned
        # and choose the type accordingly. Difficult though with those temporary downloaded files
        # "fileSystemType": "directory/archive",
        return fileentry_bdio

    @staticmethod
    def evidence_file(
        componententry: JsonDict,
        fileentry: JsonDict,
        match_type: str = "binary",
        annotation: Optional[JsonDict] = None,
    ) -> JsonDict:
        """
        Tells where a component was found, in which file.
        """
        dependency = {
            "@type": "Dependency",
            "dependsOn": componententry["@id"],
            "evidence": fileentry["@id"],
        }
        if match_type:
            dependency["matchType"] = match_type
        if annotation:
            dependency["description"] = annotation["@id"]
        return dependency

    @staticmethod
    def dependency_transitive_entry(
        dependency_component: JsonDict,
        match_type: str = "transitive dependency",
        annotation: Optional[JsonDict] = None,
    ) -> JsonDict:
        """
        Creates component links that are dependencies for another component.
        """
        dependency = {
            "@type": "Dependency",
            "dependsOn": dependency_component["@id"],
        }
        if match_type:
            dependency["matchType"] = match_type
        if annotation:
            dependency["description"] = annotation["@id"]
        return dependency

    @staticmethod
    def toplevel_graph(
        publisher: str,
        publisher_version: str,
        code_location_name: str,
        code_location_uri: Optional[str] = None,
        code_location_uri_auto: bool = False,
    ) -> JsonDict:
        """
        Top level bdio file content.
        """
        contents: List[JsonDict] = []
        result = {
            "@graph": contents,
            "@context": [
                "https://blackducksoftware.github.io/bdio/2.0.0",
                {"matchType": "https://blackducksoftware.com/hub#hasHubFileMatchType"},
            ],
            "creationDateTime": _format_datetime(datetime.utcnow()),
            "name": code_location_name,
            "publisher": f"{publisher}/{publisher_version}",
        }
        if code_location_uri:
            result["@id"] = code_location_uri
        if code_location_uri_auto:
            result["@id"] = f"urn:uuid:{uuid5(NAMESPACE_URL, code_location_name)}"
        return result

    @staticmethod
    def project(
        name: str,
        version: Optional[str],
        fileentry_project_base: JsonDict,
        description: Optional[JsonDict] = None,
    ) -> JsonDict:
        contents: List[JsonDict] = []
        project = {
            "@type": "Project",
            "@id": f"urn:uuid:{uuid4()}",
            "name": name,
            "base": fileentry_project_base["@id"],
            "dependency": contents,
        }
        if version is not None:
            project["version"] = version
        if description:
            project["description"] = description["@id"]
        return project

    @staticmethod
    def annotation(content: str) -> JsonDict:
        """
        Any note, comment etc.
        """
        annotation = {
            "@type": "Annotation",
            "@id": f"urn:uuid:{uuid4()}",
            "comment": content,
        }
        return annotation

    @staticmethod
    def external_id(
        forge: str,
        component_name: str,
        *id_parts: Optional[str],
        forge_separator_override: Optional[str] = None,
    ) -> Tuple[str, str]:
        """
        Returns a BD external-id as a tuple of (forge, identifier)
        """
        result_id_parts = [component_name]
        for id_part in id_parts:
            if id_part:
                result_id_parts.append(id_part)
        # default to slash as separator as most linux distros use that and we are dynamic in
        # discovering those from files
        if forge_separator_override:
            separator = forge_separator_override
        else:
            separator = BDIOEntries.forge_to_separator_char.get(forge, "/")
        return forge, separator.join(result_id_parts)


class BDIOBuilder:
    """
    A builder object for BDIO file contents.
    An example skeleton for creating a BDIO file. Note that bdio_file and project need to be
    created at the beginning as the other entities will be added to those:

    builder = BDIOBuilder()
    builder.bdio_file()
    builder.project()
    file = builder.fileentry_...()
    component = builder.component()
    annotation = builder.annotation()
    builder.dependency(component, file, annotation)
    ...
    return builder.get_result()
    """

    def __init__(self):
        self.toplevel_element: JsonDict = {}
        self.project_element: JsonDict = {}
        # below dicts are caches
        self.unique_component_entries: Dict[Tuple[str, str], JsonDict] = {}
        self.fileentry_by_bdio_path: Dict[str, JsonDict] = {}
        self.unique_annotations: Dict[str, JsonDict] = {}

    def _add_to_toplevel(self, entry: JsonDict) -> None:
        if not self.toplevel_element:
            raise ValueError("Toplevel element not set!")
        if "@graph" not in self.toplevel_element:
            raise ValueError("Toplevel missing '@graph' entry")

        self.toplevel_element["@graph"].append(entry)

    def _add_to_project(self, entry: JsonDict) -> None:
        if not self.project_element:
            raise ValueError("Project element not set!")
        if "dependency" not in self.project_element:
            raise ValueError("Project element missing 'dependency' entry!")

        self.project_element["dependency"].append(entry)

    def bdio_file(
        self,
        publisher: str,
        publisher_version: str,
        code_location_name: str,
        code_location_uri: Optional[str] = None,
        code_location_uri_auto: bool = False,
    ) -> JsonDict:
        """
        Top level bdio file content, also known as Graph in other places.
        """
        self.toplevel_element = BDIOEntries.toplevel_graph(
            publisher,
            publisher_version,
            code_location_name,
            code_location_uri,
            code_location_uri_auto,
        )
        return self.toplevel_element

    def project(
        self,
        name: str,
        version: str,
        fileentry_project_base: JsonDict,
        description: Optional[JsonDict] = None,
    ) -> JsonDict:
        if self.project_element:
            raise ValueError("Project element existing already. Only 1 allowed.")
        self.project_element = BDIOEntries.project(
            name, version, fileentry_project_base, description
        )
        self._add_to_toplevel(self.project_element)
        return self.project_element

    def fileentry_from_extended_object(
        self,
        extended_object: JsonDict,
        annotation: Optional[JsonDict] = None,
        original_scanned_filename: Optional[str] = None,
    ) -> JsonDict:
        extended_fullpath = extended_object.get("extended-fullpath")
        if not extended_fullpath:
            # older scan results do not have extended-fullpaths so need to mock them
            extended_fullpath = bdio_paths.simple_fullpath_to_extended_fullpath(
                extended_object["fullpath"]
            )
        filesize = extended_object.get("size")  # again to deal with older scans
        sha1 = extended_object["sha1"]
        timestamp = extended_object["timestamp"]
        return self.fileentry_from_extended_fullpath(
            extended_fullpath, timestamp, sha1, filesize, annotation, original_scanned_filename
        )

    def fileentry_from_extended_fullpath(
        self,
        extended_fullpath: ExtendedFullPath,
        timestamp: int,
        sha1: str,
        filesize: Optional[int] = None,
        annotation: Optional[JsonDict] = None,
        original_scanned_filename: Optional[str] = None,
    ) -> JsonDict:
        """
        Describe a file with its path and various other attributes.
        """
        # first file has a name in the scan project that might have been changed during the
        # upload to some temporary random file name, so patch it back here
        fileentry_bdio = BDIOEntries.fileentry(
            extended_fullpath, timestamp, sha1, filesize, annotation, original_scanned_filename
        )
        bdio_path = fileentry_bdio["path"]
        cached_fileentry_bdio = self.fileentry_by_bdio_path.get(bdio_path)
        if not cached_fileentry_bdio:
            self.fileentry_by_bdio_path[bdio_path] = fileentry_bdio
            self._add_to_toplevel(fileentry_bdio)
            cached_fileentry_bdio = fileentry_bdio
        return cached_fileentry_bdio

    def fileentry_from_fullpath(
        self,
        fullpath: List[str],
        timestamp: int,
        sha1: str,
        filesize: Optional[int] = None,
        annotation: Optional[JsonDict] = None,
        original_scanned_filename: Optional[str] = None,
    ) -> JsonDict:
        """
        Describe a file with its path and various other attributes.
        """
        extended_fullpath = bdio_paths.simple_fullpath_to_extended_fullpath(fullpath)
        return self.fileentry_from_extended_fullpath(
            extended_fullpath, timestamp, sha1, filesize, annotation, original_scanned_filename
        )

    def fileentry_project_base(self, scanned_path: str, filesize: Optional[int] = None) -> JsonDict:
        fileentry_bdio = BDIOEntries.fileentry_project_base(scanned_path, filesize)
        bdio_path = fileentry_bdio["path"]
        cached_fileentry_bdio = self.fileentry_by_bdio_path.get(bdio_path)
        if not cached_fileentry_bdio:
            self.fileentry_by_bdio_path[bdio_path] = fileentry_bdio
            self._add_to_toplevel(fileentry_bdio)
            cached_fileentry_bdio = fileentry_bdio
        return cached_fileentry_bdio

    def evidence_file(
        self,
        component_entry: JsonDict,
        fileentry: JsonDict,
        match_type: str = "binary",
        annotation: Optional[JsonDict] = None,
    ) -> JsonDict:
        """
        Tells where a component was found, in which file.

        :param component_entry: a component entry created with this builder
        :param fileentry: a file entry created with this builder that is the evidence
                            for the component
        :param match_type: one of "binary, manual, file, partial_file, snippet, dependency"
        :param annotation: any text annotation that should be linked to this
        """
        dependency = BDIOEntries.evidence_file(component_entry, fileentry, match_type, annotation)
        self._add_to_project(dependency)
        return dependency

    def dependency_direct(
        self,
        component_entry: JsonDict,
    ) -> JsonDict:
        """
        Just a component entry, without link to any evidence. Usually from a package manager.
        """
        # a direct dependency is just a component without any other references to it, e.g.
        # files or other components, so only thing to do is to make sure this component
        # is added to the graph
        forge = component_entry["namespace"]
        identifier = component_entry["identifier"]
        self.component_from_id(
            forge, identifier
        )  # this is a no-op if the component was already added
        return component_entry

    def dependency_direct_with_evidence_file(
        self,
        component_entry: JsonDict,
        fileentry: JsonDict,
        match_type: str = "binary",
        annotation: Optional[JsonDict] = None,
    ) -> JsonDict:
        self.dependency_direct(component_entry)
        # the duplicated component is necessary because the Hub internal magic on deciding matching
        # types and showing evidence is quite a bunch of ... and would not show the file if
        # it is associated with a direct dependency component. Hence the need to add another
        # component of the same ID which will be collated in the Hub view. This is a HACK!
        component_copy = self.component_copy(component_entry)
        return self.evidence_file(component_copy, fileentry, match_type, annotation)

    @staticmethod
    def dependency_transitive(
        component_entry: JsonDict,
        component_transitive: JsonDict,
        match_type: str = "transitive dependency",
        annotation: Optional[JsonDict] = None,
    ) -> JsonDict:
        """
        A component that is pulled in by another component as a transitive dependency.

        :param component_entry: a component entry created with this builder
        :param component_transitive: a component entry created with this builder on which the
                                        above component depends on
        :param match_type: one of "binary, manual, file, partial_file, snippet, dependency"
        :param annotation: any text annotation that should be linked to this

        """
        dependency_entry = BDIOEntries.dependency_transitive_entry(
            component_transitive, match_type, annotation
        )
        component_entry["dependency"] = component_entry.get("dependency", [])
        component_entry["dependency"].append(dependency_entry)
        return component_entry

    def dependency_transitive_with_evidence_file(
        self,
        component_entry: JsonDict,
        component_transitive: JsonDict,
        fileentry_transitive: JsonDict,
        match_type_transitive: str = "transitive dependency",
        match_type_file: str = "binary",
        annotation: Optional[JsonDict] = None,
    ) -> JsonDict:
        self.dependency_transitive(
            component_entry, component_transitive, match_type_transitive, annotation
        )
        # the duplicated component is necessary because the Hub internal magic on deciding matching
        # types and showing evidence is quite a bunch of ... and would not show the file if
        # it is associated with a direct dependency component. Hence the need to add another
        # component of the same ID which will be collated in the Hub view. This is a HACK!
        component_copy = self.component_copy(component_transitive)
        return self.evidence_file(component_copy, fileentry_transitive, match_type_file, annotation)

    def component_from_id(
        self, forge: str, identifier: str, fallback_forge: bool = False, allow_duplicates=False
    ) -> JsonDict:
        component = self.unique_component_entries.get((forge, identifier))
        if not component or allow_duplicates:
            # if we add a duplicate here then that will be the one returned next time
            component = BDIOEntries.component(forge, identifier, fallback_forge=fallback_forge)
            self.unique_component_entries[(forge, identifier)] = component
            self._add_to_toplevel(component)
        return component

    def component(
        self,
        forge: str,
        component_name: str,
        *id_parts: Optional[str],
        forge_separator_override: Optional[str] = None,
        fallback_forge: bool = False,
        allow_duplicates=False,
    ) -> JsonDict:
        forge, identifier = BDIOEntries.external_id(
            forge, component_name, *id_parts, forge_separator_override=forge_separator_override
        )
        return self.component_from_id(
            forge, identifier, fallback_forge=fallback_forge, allow_duplicates=allow_duplicates
        )

    def component_copy(self, component: JsonDict) -> JsonDict:
        forge = component["namespace"]
        identifier = component["identifier"]
        return self.component_from_id(forge, identifier, allow_duplicates=True)

    def tool_comment(self) -> JsonDict:
        """
        Appcheck was here!
        """
        return self.annotation("Here be Binary Ducks and Dragons!")

    def matching_info(self, extended_object: JsonDict) -> JsonDict:
        """
        An annotation containing some more information bits from our matching
        to help with debugging later.
        """
        # some of the fields below did not always exist so be careful with older results
        matching_methods = extended_object.get("matching-methods")
        if not matching_methods:
            matching_method = extended_object.get("matching-method")
            matching_methods = [matching_method] if matching_method else []
        matching_methods_string = (
            f" matching-methods: {matching_methods};" if matching_methods else ""
        )

        codetype = extended_object.get("type")
        codetype_string = f" codetype: {codetype};" if codetype else ""
        confidence = extended_object.get("confidence")
        confidence_string = f"matching-confidence: {confidence};" if confidence else ""

        return self.annotation(f"{confidence_string}{matching_methods_string}{codetype_string}")

    def annotation(self, content: str) -> JsonDict:
        """
        Any note, comment etc.
        """
        annotation = self.unique_annotations.get(content)
        if not annotation:
            annotation = BDIOEntries.annotation(content)
            self.unique_annotations[content] = annotation
            self._add_to_toplevel(annotation)
        return annotation

    def get_result(self) -> JsonDict:
        return self.toplevel_element



"""
Copyright (c) 2021 Synopsys, Inc.
Use subject to the terms and conditions of the Synopsys End User Software License and Maintenance Agreement.
All rights reserved worldwide.
"""
import requests
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
import urllib3
import logging
import os
import tqdm

# Disable warnings: InsecureRequestWarning
urllib3.disable_warnings()


def read_in_chunks(file_object_1, chunk_size=1024000000):
    """Generator to read a file in chunks.
    Default chunk size: 1GB."""
    while True:
        data = file_object_1.read(chunk_size)
        if not data:
            break
        yield data


class BDBAApi:
    """
    Class that facilitates an authenticated connection to the hub for uploading binary files to BDBA.
    """

    def __init__(self, hub_api):
        hub_api.authenticate()
        self.hub = hub_api.hub

    def upload_binary_part(self, project_name, version_name, codelocation_name, binary_file, file_chunk):
        """ upload bdba tar file to hub"""
        if not os.path.exists(binary_file):
            logging.warning("Binary file not found. There will be no BDBA results.")
            return

        multipart_encoder = MultipartEncoder(fields={'projectName': project_name,
                                                     'version': version_name,
                                                     'codeLocationName': codelocation_name,
                                                     'fileupload': (binary_file, file_chunk, 'application/octet-stream')
                                                     }
                                             )
        headers = self.hub.get_headers()
        headers.pop('Accept')
        headers.pop('Content-Type')

        with tqdm.tqdm(desc="BDBA Upload",
                       total=multipart_encoder.len,
                       dynamic_ncols=True,
                       unit='B',
                       unit_scale=True,
                       unit_divisor=1024) as bar:
            multipart_monitor = MultipartEncoderMonitor(multipart_encoder,
                                                        lambda monitor: bar.update(monitor.bytes_read - bar.n))
            headers['Content-Type'] = multipart_monitor.content_type
            try:
                response = requests.post('{}/api/uploads'.format(self.hub.config['baseurl']),
                                         headers=headers,
                                         data=multipart_monitor,
                                         verify=not self.hub.config['insecure'], timeout=600)
                if not response.ok:
                    if response.status_code == 402:
                        logging.error("BDBA is not licensed for use with the current Black Duck instance -- will not be used.")
                    else:
                        logging.error(
                            "Problem uploading the file to BDBA -- (Response({}): {})".format(response.status_code,
                                                                                              response.text))
                else:
                    logging.info("Upload completed!")

            except requests.exceptions.ReadTimeout as e:
                logging.error("Timeout error occurred when uploading BDBA file {}".format(e))

    def upload_binary(self, project_name, version_name, codelocation_name, binary_file):
        """send chunk of binary file to upload"""
        if not os.path.exists(binary_file):
            logging.warning("Binary file not found. There will be no BDBA results.")
            return
        bdba_count = 0
        with open(binary_file, 'rb') as file_obj:
            for file_chunk in read_in_chunks(file_obj):
                bdba_count += 1
                logging.info("Uploading bdba tar file part {}".format(bdba_count))
                upd_codelocation_name = codelocation_name + "-" + str(bdba_count) + "-BDBA"
                self.upload_binary_part(project_name, version_name, upd_codelocation_name, binary_file, file_chunk)

    def __call__(self, project_name, version_name, codelocation_name, binary_file):
        try:
            self.upload_binary(project_name, version_name, codelocation_name, binary_file)
        except MemoryError as e:
            logging.error("Please verify if you are using right version of python - 32bit vs 64bit - {}".format(e))



"""
Copyright (c) 2021 Synopsys, Inc.
Use subject to the terms and conditions of the Synopsys End User Software License and Maintenance Agreement.
All rights reserved worldwide.
"""

from typing import Dict

from blackduck_c_cpp.katiska.bdio.bdio_entries import (
    BDIOBuilder,
)

from blackduck_c_cpp.katiska.custom_types import JsonDict

from typing import List
import json
import os

'''
Python wrapper to convert json to bdio2 format 
@author Mayur kadu
@date March 11 2021
'''


class BDIO2Transformer:
    BDIO2_OUT_FILENAME = "c_cpp_bdio2.jsonld"
    PUBLISHER = "c-cpp-bdio2"
    PUBLISHER_VERSION = "1.0.0"
    BDIO2_FILENAME_SUFFIX = "-BDIO2"

    def __init__(self, cov_output, hub_api):
        self.cov_output = cov_output
        self.out_biod2_filepath_name = os.path.join(self.cov_output, self.BDIO2_OUT_FILENAME)
        self.hub_api = hub_api
        self.vers_result = self.hub_api.hub.version_info['version'].split(".")


    def write_bdio2_document(self, result):
        '''
        Write the bdio2 file to cov_output path
        @return : path with filename of bdio2 file generated
        '''
        with open(self.out_biod2_filepath_name, "w") as outfile:
            json.dump(result, outfile, indent=2)

        return self.out_biod2_filepath_name

    def generate_bdio2(self, code_loc_name,
                       project_name,
                       project_version,
                       cov_results: List[JsonDict],
                       scanned_path,
                       scan_size=0
                       ):
        '''
        code_loc_name : Code location name
        project_name  : Project Name
        project_version : project version
        cov_result : List[JsonDict] components identified
        scanned_path : coverity output path of project
        scan_size : size of scan

        @return : string representation of jsonld bdio2
        '''

        code_location_name = code_loc_name + self.BDIO2_FILENAME_SUFFIX
        if self.vers_result[0] == 2021 and self.vers_result[1] == 8:
            self.PUBLISHER = "Protecode-SC"
            self.PUBLISHER_VERSION = "1.0.0"
        bdio_builder = BDIOBuilder()
        bdio_builder.bdio_file(self.PUBLISHER, self.PUBLISHER_VERSION, code_location_name, None, True)

        fileentry_project_base = bdio_builder.fileentry_project_base(scanned_path, scan_size)
        bdio_builder.project(project_name, project_version, fileentry_project_base)
        dir_comp_dict = dict()  # key: direct dependency compoenent ,value:(comp value, set of fullpath)

        for entry in cov_results[0]["extended-objects"]:
            # all entrys are direct deps
            comp = entry["distro"] + entry["package-name"] + entry["package-version"] + entry["package-architecture"]

            if comp in dir_comp_dict and entry['fullpath'][1] in dir_comp_dict[comp][1]:
                continue

            if comp in dir_comp_dict:  # use existing direct dependency component
                direct_dep_component = dir_comp_dict[comp][0]
                dir_comp_dict[comp][1].add(entry['fullpath'][1])  # append path
            else:  # else create new one if it was never seen before
                direct_dep_component = bdio_builder.component(
                    entry["distro"],
                    entry["package-name"],
                    entry["package-version"],
                    entry["package-architecture"],
                    allow_duplicates=True,
                )
                dir_comp_dict[comp] = (direct_dep_component, {entry['fullpath'][1]})

            direct_dep_file = bdio_builder.fileentry_from_fullpath(
                entry["fullpath"],
                entry["timestamp"],
                entry["sha1"],
                entry["size"],
                bdio_builder.matching_info(entry),
            )
            bdio_builder.dependency_direct_with_evidence_file(
                direct_dep_component, direct_dep_file, match_type=entry["matchType"]
            )

            # add all transitive components
            transitive_entry: Dict = {}  # needed for type checker to stop complaining
            for transitive_entry in entry.get("trans-dep", []):
                transitive_dep_component = bdio_builder.component(
                    transitive_entry["distro"],
                    transitive_entry["package-name"],
                    transitive_entry["package-version"],
                    transitive_entry["package-architecture"],
                    allow_duplicates=True,
                )
                transitive_dep_file = bdio_builder.fileentry_from_fullpath(
                    transitive_entry["fullpath"],
                    transitive_entry["timestamp"],
                    transitive_entry["sha1"],
                    transitive_entry["size"],
                    bdio_builder.matching_info(transitive_entry),
                )
                bdio_builder.dependency_transitive_with_evidence_file(
                    direct_dep_component,
                    transitive_dep_component,
                    transitive_dep_file,
                    match_type_file=transitive_entry["matchType"],
                )
        bdio = bdio_builder.get_result()

        return bdio, code_location_name



"""
Copyright (c) 2021 Synopsys, Inc.
Use subject to the terms and conditions of the Synopsys End User Software License and Maintenance Agreement.
All rights reserved worldwide.
"""

import subprocess
import logging
import os
import sys
import glob

'''
Utility class to keep commonly used methods/functionality here
'''


def conf_cmd(spec, cov_bin, cov_conf):
    try:
        cov_configure_path = glob.glob(os.path.join(cov_bin, 'cov-configure*'))[0]
    except IndexError:
        error_and_exit.error("cov-configure not found in bin directory at location: {}".format(cov_bin))
    return '"{}" -c "{}" {}'.format(cov_configure_path, os.path.join(cov_conf, "bld.xml"), spec)


def run_cmd(cmd, curdir=None, suppress_output=False, env=None):
    logging.debug("command to run in run_cmd {}".format(cmd))
    if env and curdir:
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env, shell=True,
                                   cwd=curdir)
    elif env:
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env, shell=True)
    elif curdir:
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True, cwd=curdir)
    else:
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True)
    while True:
        try:
            realtime_output = process.stdout.readline().decode("utf-8")
            if realtime_output == '' and process.poll() is not None:
                break
            if realtime_output:
                try:
                    if not suppress_output:
                        logging.info(realtime_output.strip())
                    if suppress_output:
                        logging.debug(realtime_output.strip())
                except UnicodeEncodeError:
                    logging.warning("UnicodeEncodeError Exception occurred: skipping a line")
                except UnicodeDecodeError:
                    logging.warning("UnicodeDecodeError Exception occurred: skipping a line")
        except UnicodeEncodeError:
            logging.warning("UnicodeEncodeError Exception occurred: skipping a line")
        except UnicodeDecodeError:
            logging.warning("UnicodeDecodeError Exception occurred: skipping a line")
    return process.returncode


def run_cmd_emit(cmd):
    logging.debug("command to run in run_cmd_emit {}".format(cmd))
    p = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True, encoding='utf-8')
    return p.stdout.replace(os.linesep, '\n').split('\n')


def value_getstatusoutput(output):
    status = False
    if output[0] == 0:  ## success
        status = True
        return status, output[1]
    else:  ## failed
        return status, "no path found matching pattern"


def resolve_path(path):
    """
    Function to resolve a path that contains "../" segments
    For example /lib/a/b/../c -> /lib/a/c.
    param: (string) path
    return: (string) resolved path
    """
    new_path = []
    for component in path.split('/'):
        if len(component) > 0:
            if component == '..':
                new_path = new_path[0:len(new_path) - 1]
            else:
                new_path = new_path + [component]
    return '/' + '/'.join(new_path)


def error_and_exit(message):
    logging.error(message)
    sys.exit(1)



"""
Copyright (c) 2021 Synopsys, Inc.
Use subject to the terms and conditions of the Synopsys End User Software License and Maintenance Agreement.
All rights reserved worldwide.
"""

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="blackduck-c-cpp",
    version="1.0.4",
    description="Scanning for c/c++ projects using blackduck and coverity tools",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=["requests>=2.5.3",
                      "numpy>=1.19.5",
                      "pandas>=1.1.5",
                      "python-dateutil>=2.8.1",
                      "pytz>=2021.1",
                      "six>=1.15.0",
                      "tqdm>=4.58.0",
                      "blackduck>=1.0.4",
                      "configargparse>=1.4",
                      "structlog>=20.1.0",
                      "pyyaml>=5.3",
                      "urllib3>=1.22",
                      "requests-toolbelt>=0.9.0",
                      "google-cloud-storage>=1.42.3"
                      ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Other/Proprietary License",
    ],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': ['blackduck-c-cpp=blackduck_c_cpp.run_build_capture:run']},
)
